from typing import List
from typing import Optional
from typing import Union

from tecton._internals import errors
from tecton._internals import metadata_service
from tecton._internals.display import Displayable
from tecton._internals.sdk_decorators import documented_by
from tecton._internals.sdk_decorators import sdk_public_method
from tecton._internals.utils import format_freshness_table
from tecton._internals.utils import get_all_freshness
from tecton._internals.utils import is_live_workspace
from tecton.framework import data_source as framework_data_source
from tecton.framework import entity as framework_entity
from tecton.framework import feature_service as framework_feature_service
from tecton.framework import feature_view as framework_feature_view
from tecton.framework import transformation as framework_transformation
from tecton.framework.dataset import Dataset
from tecton_core import specs
from tecton_core.fco_container import FcoContainer
from tecton_core.feature_definition_wrapper import FrameworkVersion
from tecton_core.logger import get_logger
from tecton_proto.metadataservice.metadata_service_pb2 import GetAllEntitiesRequest
from tecton_proto.metadataservice.metadata_service_pb2 import GetAllFeatureServicesRequest
from tecton_proto.metadataservice.metadata_service_pb2 import GetAllSavedFeatureDataFramesRequest
from tecton_proto.metadataservice.metadata_service_pb2 import GetAllTransformationsRequest
from tecton_proto.metadataservice.metadata_service_pb2 import GetAllVirtualDataSourcesRequest
from tecton_proto.metadataservice.metadata_service_pb2 import GetEntityRequest
from tecton_proto.metadataservice.metadata_service_pb2 import GetFeatureServiceRequest
from tecton_proto.metadataservice.metadata_service_pb2 import GetFeatureViewRequest
from tecton_proto.metadataservice.metadata_service_pb2 import GetSavedFeatureDataFrameRequest
from tecton_proto.metadataservice.metadata_service_pb2 import GetTransformationRequest
from tecton_proto.metadataservice.metadata_service_pb2 import GetVirtualDataSourceRequest
from tecton_proto.metadataservice.metadata_service_pb2 import ListWorkspacesRequest
from tecton_proto.metadataservice.metadata_service_pb2 import QueryFeatureViewsRequest


logger = get_logger("Workspace")


class Workspace:
    """
    Workspace class.

    This class represents a Workspace. The Workspace class is used to fetch Tecton Objects, which are stored in a Workspace.

    Examples of using the Workspace methods for accessing Tecton first class objects:

    .. code-block:: python

        import tecton


        workspace = tecton.get_workspace("my_workspace")

        # For a specified workspace, list high-level registered Objects
        print(f" Entities            : {workspace.list_entities()}")
        print(f" Feature Data Sources: {workspace.list_data_sources()}")
        print(f" Feature Views       : {workspace.list_feature_views()}")
        print(f" Feature Services    : {workspace.list_feature_services()}")
        print(f" Transformations     : {workspace.list_transformations()}")
        print(f" Feature Tables      : {workspace.list_feature_tables()}")
        print(f" List of Workspaces  : {tecton.list_workspaces()}")

    .. code-block:: Text

        The output:

        Entities            : ['ad', 'auction', 'content', 'ContentKeyword', 'ads_user', 'fraud_user']
        Feature Data Sources: ['ad_impressions_stream', 'ad_impressions_batch', 'transactions_stream', 'transactions_batch',
                                'users_batch', 'credit_scores_batch']
        Feature Views       : ['user_has_great_credit', 'user_age', 'transaction_amount_is_high',
                               'transaction_amount_is_higher_than_average', 'transaction_bucketing', 'user_ctr_7d_2',
                               ...
                               'user_ad_impression_counts', 'content_keyword_click_counts']
        Feature Services    : ['ad_ctr_feature_service', 'fraud_detection_feature_service',
                                'fraud_detection_feature_service:v2', 'minimal_fs', 'continuous_feature_service']
        Transformations     : ['content_keyword_click_counts', 'user_ad_impression_counts', 'user_click_counts',
                               'user_impression_counts', 'user_ctr_7d', 'user_ctr_7d_2', 'user_distinct_ad_count_7d',
                               ...
                               'weekend_transaction_count_n_days', 'user_has_good_credit_sql']
        Feature Tables      : ['user_login_counts', 'user_page_click_feature_table']
        List of Workspaces  : ['jsd_tecton_wksp, 'kafka_streaming_staging, 'kafka_streaming_production',
                               ...
                               'on_demand_streaming_aggregation_pipeline']

    """

    def __init__(self, workspace: str, _is_live: Optional[bool] = None, _validate: bool = True):
        """
        Fetch an existing :class:`tecton.Workspace` by name.

        :param workspace: Workspace name.
        """
        self.workspace = workspace

        if _is_live is None:
            self.is_live = is_live_workspace(self.workspace)
        else:
            self.is_live = _is_live

        if _validate:
            self._validate()

    def _validate(self):
        request = ListWorkspacesRequest()
        response = metadata_service.instance().ListWorkspaces(request)

        workspace_from_resp = None
        for ws in response.workspaces:
            if ws.name == self.workspace:
                workspace_from_resp = ws
                break

        if workspace_from_resp is None:
            raise errors.NONEXISTENT_WORKSPACE(self.workspace, response.workspaces)

        if ws.capabilities.materializable != self.is_live:
            raise errors.INCORRECT_MATERIALIZATION_ENABLED_FLAG(self.is_live, ws.capabilities.materializable)

    @classmethod
    @sdk_public_method
    # TODO(TEC-13678): consolidate with tecton.list_workspaces() and remove this
    def get_all(self) -> List["Workspace"]:
        """Deprecated. Use tecton.list_workspaces() instead. Returns a list of all registered Workspaces.

        :return: A list of Workspace objects.
        """
        logger.warn("get_all() is deprecated. Use tecton.list_workspaces() instead.")
        request = ListWorkspacesRequest()
        response = metadata_service.instance().ListWorkspaces(request)
        workspaces = [
            Workspace(ws.name, _is_live=ws.capabilities.materializable, _validate=False) for ws in response.workspaces
        ]

        # Return live workspaces first (alphabetical), then development workspaces.
        return sorted(workspaces, key=lambda ws: (not ws.is_live, ws.workspace))

    def __repr__(self) -> str:
        capability_str = "Live" if self.is_live else "Development"
        return f"{self.workspace} ({capability_str})"

    @sdk_public_method
    def summary(self) -> Displayable:
        """Returns workspace metadata.

        :return: Displayable of metadata for this workspace. Use `to_dict()` for a parseable representation.
        """
        items = [
            ("Workspace Name", self.workspace),
            ("Workspace Type", "Live" if self.is_live else "Development"),
        ]
        return Displayable.from_properties(items=items)

    @classmethod
    @sdk_public_method
    def get(cls, name) -> "Workspace":
        """Returns a :class:`tecton.Workspace` instance for the workspace with the provided name.

        :param name: The name of the workspace to retrieve.
        """
        return Workspace(name)

    @sdk_public_method
    def get_feature_view(self, name: str) -> framework_feature_view.FeatureView:
        """Returns a Feature View that has been applied to a workspace.

        :param name: The name of the Feature View to retrieve.
        """
        request = GetFeatureViewRequest(version_specifier=name, workspace=self.workspace)
        response = metadata_service.instance().GetFeatureView(request)
        fco_container = FcoContainer.from_proto(response.fco_container)
        feature_view_spec = fco_container.get_single_root()

        if feature_view_spec is None:
            raise errors.TectonValidationError(
                f"Feature View '{name}' not found. Try running `workspace.list_feature_views()` to view all registered Feature Views."
            )

        assert isinstance(feature_view_spec, specs.FeatureViewSpec)

        if isinstance(feature_view_spec, specs.FeatureTableSpec):
            raise errors.TectonValidationError(
                f"Feature View '{name}' not found. Did you mean workspace.get_feature_table('{name}')?"
            )

        return framework_feature_view.feature_view_from_spec(feature_view_spec, fco_container)

    @sdk_public_method
    def get_feature_table(self, name: str) -> framework_feature_view.FeatureTable:
        """Returns a Feature Table that has been applied to a workspace.

        :param name: The name of the Feature Table to retrieve.
        """
        request = GetFeatureViewRequest(version_specifier=name, workspace=self.workspace)
        response = metadata_service.instance().GetFeatureView(request)
        fco_container = FcoContainer.from_proto(response.fco_container)
        feature_table_spec = fco_container.get_single_root()

        if feature_table_spec is None:
            raise errors.TectonValidationError(
                f"Feature Table '{name}' not found. Try running `workspace.list_feature_tables()` to view all registered Feature Tables."
            )

        assert isinstance(feature_table_spec, specs.FeatureViewSpec)
        ft_proto = feature_table_spec.data_proto

        if not ft_proto.HasField("feature_table"):
            raise errors.TectonValidationError(
                f"Feature Table '{name}' not found. Did you mean workspace.get_feature_view('{name}')?"
            )

        if ft_proto.fco_metadata.framework_version != FrameworkVersion.FWV5.value:
            raise errors.UNSUPPORTED_FRAMEWORK_VERSION

        return framework_feature_view.FeatureTable._from_spec(feature_table_spec, fco_container)

    @sdk_public_method
    def get_feature_service(self, name: str) -> framework_feature_service.FeatureService:
        """Returns a Feature Service that has been applied to a workspace.

        :param name: The name of the Feature Service to retrieve.
        """
        request = GetFeatureServiceRequest(service_reference=name, workspace=self.workspace)
        response = metadata_service.instance().GetFeatureService(request)
        fco_container = FcoContainer.from_proto(response.fco_container)
        feature_service_spec = fco_container.get_single_root()

        if feature_service_spec is None:
            raise errors.TectonValidationError(
                f"Feature Service '{name}' not found. Try running `workspace.list_feature_services()` to view all registered Feature Services."
            )

        assert isinstance(feature_service_spec, specs.FeatureServiceSpec)

        return framework_feature_service.FeatureService._from_spec(feature_service_spec, fco_container)

    @sdk_public_method
    def get_data_source(
        self, name: str
    ) -> Union[
        framework_data_source.PushSource,
        framework_data_source.BatchSource,
        framework_data_source.StreamSource,
    ]:
        """Returns a Data Source that has been applied to a workspace.

        :param name: The name of the Data Source to retrieve.
        """
        request = GetVirtualDataSourceRequest(name=name, workspace=self.workspace)
        response = metadata_service.instance().GetVirtualDataSource(request)
        fco_container = FcoContainer.from_proto(response.fco_container)
        data_source_spec = fco_container.get_single_root()

        if data_source_spec is None:
            raise errors.TectonValidationError(
                f"Data Source '{name}' not found. Try running `workspace.list_data_sources()` to view all registered Data Sources."
            )

        assert isinstance(data_source_spec, specs.DataSourceSpec)
        ds_proto = data_source_spec.data_proto

        if ds_proto.fco_metadata.framework_version != FrameworkVersion.FWV5.value:
            raise errors.UNSUPPORTED_FRAMEWORK_VERSION

        return framework_data_source.data_source_from_spec(data_source_spec)

    @sdk_public_method
    def get_entity(self, name: str) -> framework_entity.Entity:
        """Returns an Entity that has been applied to a workspace.

        :param name: The name of the Entity to retrieve.
        """

        request = GetEntityRequest(name=name, workspace=self.workspace)
        response = metadata_service.instance().GetEntity(request)

        fco_container = FcoContainer.from_proto(response.fco_container)
        entity_spec = fco_container.get_single_root()

        if entity_spec is None:
            raise errors.TectonValidationError(
                f"Entity '{name}' not found. Try running `workspace.list_entities()` to view all registered Entities."
            )

        assert isinstance(entity_spec, specs.EntitySpec)

        return framework_entity.Entity._from_spec(entity_spec)

    @sdk_public_method
    def get_transformation(self, name: str) -> framework_transformation.Transformation:
        """Returns a Transformation that has been applied to a workspace.

        :param name: The name of the Transformation to retrieve.
        """
        request = GetTransformationRequest(name=name, workspace=self.workspace)
        response = metadata_service.instance().GetTransformation(request)
        fco_container = FcoContainer.from_proto(response.fco_container)
        transformation_spec = fco_container.get_single_root()

        if transformation_spec is None:
            raise errors.TectonValidationError(
                f"Transformation '{name}' not found. Try running `workspace.list_transformations()` to view all registered Transformations."
            )

        assert isinstance(transformation_spec, specs.TransformationSpec)
        return framework_transformation.Transformation._from_spec(transformation_spec)

    @sdk_public_method
    def get_dataset(self, name: str) -> Dataset:
        """Returns a Dataset that has been saved to this workspace.

        :param name: The name of the Dataset to retrieve.
        """
        request = GetSavedFeatureDataFrameRequest(saved_feature_dataframe_name=name, workspace=self.workspace)
        response = metadata_service.instance().GetSavedFeatureDataFrame(request)

        return Dataset._from_proto(response.saved_feature_dataframe)

    @sdk_public_method
    def list_datasets(self) -> List[str]:
        """Returns a list of all saved Datasets within a workspace.

        :return: List of strings of dataset names"""
        request = GetAllSavedFeatureDataFramesRequest(workspace=self.workspace)
        response = metadata_service.instance().GetAllSavedFeatureDataFrames(request)
        return sorted([sfdf.info.name for sfdf in response.saved_feature_dataframes])

    @sdk_public_method
    def list_feature_views(self) -> List[str]:
        """Returns a list of all registered Feature Views within a workspace.

        :return: List of strings of feature view names"""
        request = QueryFeatureViewsRequest(workspace=self.workspace)
        response = metadata_service.instance().QueryFeatureViews(request)
        fco_container = FcoContainer.from_proto(response.fco_container)
        return sorted(
            [spec.name for spec in fco_container.get_root_fcos() if not isinstance(spec, specs.FeatureTableSpec)]
        )

    @sdk_public_method
    def list_feature_services(self) -> List[str]:
        """Returns a list of all registered Feature Services within a workspace.

        :return: List of strings of feature service names"""
        request = GetAllFeatureServicesRequest()
        request.workspace = self.workspace
        response = metadata_service.instance().GetAllFeatureServices(request)
        return sorted([proto.fco_metadata.name for proto in response.feature_services])

    @sdk_public_method
    def list_transformations(self) -> List[str]:
        """Returns a list of all registered Transformations within a workspace.

        :return: List of strings of transformation names"""
        request = GetAllTransformationsRequest(workspace=self.workspace)
        response = metadata_service.instance().GetAllTransformations(request)
        return sorted([proto.fco_metadata.name for proto in response.transformations])

    @sdk_public_method
    def list_entities(self) -> List[str]:
        """Returns a list of all registered Entities within a workspace.

        :return: List of strings of entity names"""
        request = GetAllEntitiesRequest(workspace=self.workspace)
        response = metadata_service.instance().GetAllEntities(request)
        return sorted([proto.fco_metadata.name for proto in response.entities])

    @sdk_public_method
    def list_data_sources(self) -> List[str]:
        """Returns a list of all registered Data Sources within a workspace.

        :return: List of strings of data source names"""
        request = GetAllVirtualDataSourcesRequest(workspace=self.workspace)
        response = metadata_service.instance().GetAllVirtualDataSources(request)
        return sorted([proto.fco_metadata.name for proto in response.virtual_data_sources])

    @sdk_public_method
    def list_feature_tables(self) -> List[str]:
        """Returns a list of all registered Feature Tables within a workspace.

        :return: List of strings of feature table names"""
        request = QueryFeatureViewsRequest(workspace=self.workspace)
        response = metadata_service.instance().QueryFeatureViews(request)
        fco_container = FcoContainer.from_proto(response.fco_container)
        return sorted([spec.name for spec in fco_container.get_root_fcos() if isinstance(spec, specs.FeatureTableSpec)])

    @sdk_public_method
    def get_feature_freshness(self) -> Union[Displayable, str]:
        """Returns feature freshness status for Feature Views and Tables.

        :return: Displayable containing freshness statuses for all features. Use `to_dict()` for a parseable representation.
        """
        freshness_statuses = get_all_freshness(self.workspace)

        if len(freshness_statuses) == 0:
            return "No Feature Views found in this workspace"
        return format_freshness_table(freshness_statuses)


@sdk_public_method
@documented_by(Workspace.get)
def get_workspace(name: str):
    return Workspace.get(name)


@sdk_public_method
# Deprecated methods - kept around so that there is a helpful error message. Can be cleaned up after the 0.6 cut.
def get_feature_service(name: str, workspace_name: Optional[str] = None):
    raise errors.TectonValidationError(
        'get_feature_service must be called from a Workspace object. E.g. tecton.get_workspace("<workspace>").get_feature_service("<feature service>").'
    )


@sdk_public_method
# Deprecated methods - kept around so that there is a helpful error message. Can be cleaned up after the 0.6 cut.
def get_feature_table(ft_reference: str, workspace_name: Optional[str] = None):
    raise errors.TectonValidationError(
        'get_feature_table must be called from a Workspace object. E.g. tecton.get_workspace("<workspace>").get_feature_table("<feature table>").'
    )


@sdk_public_method
# Deprecated methods - kept around so that there is a helpful error message. Can be cleaned up after the 0.6 cut.
def get_feature_view(fv_reference: str, workspace_name: Optional[str] = None):
    raise errors.TectonValidationError(
        'get_feature_view must be called from a Workspace object. E.g. tecton.get_workspace("<workspace>").get_feature_view("<feature view>").'
    )


@sdk_public_method
# Deprecated methods - kept around so that there is a helpful error message. Can be cleaned up after the 0.6 cut.
def get_entity(name: str, workspace_name: Optional[str] = None):
    raise errors.TectonValidationError(
        'get_entity must be called from a Workspace object. E.g. tecton.get_workspace("<workspace>").get_entity("<entity>").'
    )


@sdk_public_method
# Deprecated methods - kept around so that there is a helpful error message. Can be cleaned up after the 0.6 cut.
def get_transformation(name, workspace_name: Optional[str] = None):
    raise errors.TectonValidationError(
        'get_transformation must be called from a Workspace object. E.g. tecton.get_workspace("<workspace>").get_transformation("<transformation>").'
    )


@sdk_public_method
# Deprecated methods - kept around so that there is a helpful error message. Can be cleaned up after the 0.6 cut.
def get_data_source(name, workspace_name: Optional[str] = None):
    raise errors.TectonValidationError(
        'get_data_source must be called from a Workspace object. E.g. tecton.get_workspace("<workspace>").get_data_source("<data source>").'
    )



from datetime import datetime
from typing import Dict
from typing import Optional

import attrs
import pendulum
import pyspark

from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper
from tecton_spark import data_observability
from tecton_spark.partial_aggregations import construct_partial_time_aggregation_df
from tecton_spark.pipeline_helper import _PipelineBuilder
from tecton_spark.pipeline_helper import dataframe_with_input
from tecton_spark.query.node import SparkExecNode
from tecton_spark.schema_spark_utils import schema_to_spark


@attrs.frozen
class OdfvPipelineSparkNode(SparkExecNode):
    input_node: SparkExecNode
    feature_definition_wrapper: FeatureDefinitionWrapper
    namespace: str

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        return dataframe_with_input(
            spark,
            self.feature_definition_wrapper.pipeline,
            self.input_node.to_dataframe(spark),
            schema_to_spark(self.feature_definition_wrapper.view_schema),
            transformations=self.feature_definition_wrapper.transformations,
            name=self.feature_definition_wrapper.name,
            fv_id=self.feature_definition_wrapper.id,
            namespace_separator=self.feature_definition_wrapper.namespace_separator,
            namespace=self.namespace,
        )


@attrs.frozen
class PipelineEvalSparkNode(SparkExecNode):
    inputs_map: Dict[str, SparkExecNode]
    feature_definition_wrapper: FeatureDefinitionWrapper

    # Needed for correct behavior by tecton_sliding_window udf if it exists in the pipeline
    feature_time_limits: Optional[pendulum.Period]

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        df = _PipelineBuilder(
            spark,
            self.feature_definition_wrapper.pipeline,
            consume_streaming_data_sources=False,
            data_sources=self.feature_definition_wrapper.data_sources,
            transformations=self.feature_definition_wrapper.transformations,
            feature_time_limits=self.feature_time_limits,
            schedule_interval=self.feature_definition_wrapper.batch_materialization_schedule,
            passed_in_inputs={k: self.inputs_map[k].to_dataframe(spark) for k in self.inputs_map},
        ).get_dataframe()
        return df


@attrs.frozen
class PartialAggSparkNode(SparkExecNode):
    input_node: SparkExecNode
    fdw: FeatureDefinitionWrapper = attrs.field()
    window_start_column_name: str
    window_end_column_name: Optional[str] = None
    aggregation_anchor_time: Optional[datetime] = None

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        df = construct_partial_time_aggregation_df(
            self.input_node.to_dataframe(spark),
            list(self.fdw.join_keys),
            self.fdw.trailing_time_window_aggregation,
            self.fdw.get_feature_store_format_version,
            window_start_column_name=self.window_start_column_name,
            window_end_column_name=self.window_end_column_name,
            aggregation_anchor_time=self.aggregation_anchor_time,
        )
        return df


@attrs.frozen
class MetricsCollectorSparkNode(SparkExecNode):
    input_node: SparkExecNode
    metrics_collector: data_observability.MetricsCollector = attrs.field(
        factory=data_observability.get_active_metrics_collector
    )

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        return self.metrics_collector.observe(self.input_node.to_dataframe(spark))



import os
import sys

from pyspark.sql import SparkSession

from tecton import conf
from tecton_core.logger import get_logger
from tecton_spark.udf_jar import get_udf_jar_path


# Environment variables to propagate to the executor
EXECUTOR_ENV_VARIABLES = [
    "AWS_ACCESS_KEY_ID",
    "AWS_SECRET_ACCESS_KEY",
    "HIVE_METASTORE_HOST",
    "HIVE_METASTORE_PORT",
    "HIVE_METASTORE_USERNAME",
    "HIVE_METASTORE_DATABASE",
    "HIVE_METASTORE_PASSWORD",
]


logger = get_logger("SparkUtils")


def _get_basic_spark_session_builder():
    builder = SparkSession.builder

    if conf.get_or_none("AWS_ACCESS_KEY_ID") is not None:
        builder = builder.config("spark.hadoop.fs.s3a.access.key", conf.get_or_none("AWS_ACCESS_KEY_ID"))
        builder = builder.config("spark.hadoop.fs.s3a.secret.key", conf.get_or_none("AWS_SECRET_ACCESS_KEY"))
    if conf.get_or_none("HIVE_METASTORE_HOST") is not None:
        builder = (
            builder.enableHiveSupport()
            .config(
                "spark.hadoop.javax.jdo.option.ConnectionURL",
                f"jdbc:mysql://{conf.get_or_none('HIVE_METASTORE_HOST')}:{conf.get_or_none('HIVE_METASTORE_PORT')}/{conf.get_or_none('HIVE_METASTORE_DATABASE')}",
            )
            .config("spark.hadoop.javax.jdo.option.ConnectionDriverName", "com.mysql.cj.jdbc.Driver")
            .config("spark.hadoop.javax.jdo.option.ConnectionUserName", conf.get_or_none("HIVE_METASTORE_USERNAME"))
            .config("spark.hadoop.javax.jdo.option.ConnectionPassword", conf.get_or_none("HIVE_METASTORE_PASSWORD"))
        )
    for v in EXECUTOR_ENV_VARIABLES:
        if conf.get_or_none(v) is not None:
            builder = builder.config(f"spark.executorEnv.{v}", conf.get_or_none(v))
    return builder


def get_or_create_spark_session(custom_spark_options=None):
    logger.debug("Creating Apache Spark Context")
    builder = _get_basic_spark_session_builder()

    if conf.get_or_none("SPARK_DRIVER_LOCAL_IP"):
        # Sadly, Spark will always attempt to retrieve the driver's local IP
        # even if the "spark.driver.host" conf is set. This will
        # result in an error if Spark's code is unable to retrieve the local IP
        # (which happens in the case of SageMaker notebooks). The only way to override Spark's
        # behavior is by setting the SPARK_LOCAL_IP env variable
        os.environ["SPARK_LOCAL_IP"] = conf.get_or_none("SPARK_DRIVER_LOCAL_IP")

    builder = builder.master("local")

    builder = builder.config("spark.jars", get_udf_jar_path())

    # For some reason using the property spark.pyspark.python does not
    # work for this like it should
    os.environ["PYSPARK_PYTHON"] = sys.executable

    if custom_spark_options:
        for k, v in custom_spark_options.items():
            builder.config(k, v)

    return builder.getOrCreate()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/args/pipeline.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from tecton_proto.common import id_pb2 as tecton__proto_dot_common_dot_id__pb2
from tecton_proto.common import spark_schema_pb2 as tecton__proto_dot_common_dot_spark__schema__pb2
from tecton_proto.args import feature_service_pb2 as tecton__proto_dot_args_dot_feature__service__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n tecton_proto/args/pipeline.proto\x12\x11tecton_proto.args\x1a\x1egoogle/protobuf/duration.proto\x1a\x1bgoogle/protobuf/empty.proto\x1a\x1ctecton_proto/common/id.proto\x1a&tecton_proto/common/spark_schema.proto\x1a\'tecton_proto/args/feature_service.proto\"z\n\x08Pipeline\x12\x33\n\x04root\x18\x01 \x01(\x0b\x32\x1f.tecton_proto.args.PipelineNodeR\x04root\x12\x33\n\x04mode\x18\x02 \x01(\x0e\x32\x1f.tecton_proto.args.PipelineModeR\x04modeJ\x04\x08\x03\x10\x04\"\xbc\x04\n\x0cPipelineNode\x12X\n\x13transformation_node\x18\x01 \x01(\x0b\x32%.tecton_proto.args.TransformationNodeH\x00R\x12transformationNode\x12M\n\x10\x64\x61ta_source_node\x18\x02 \x01(\x0b\x32!.tecton_proto.args.DataSourceNodeH\x00R\x0e\x64\x61taSourceNode\x12\x46\n\rconstant_node\x18\x03 \x01(\x0b\x32\x1f.tecton_proto.args.ConstantNodeH\x00R\x0c\x63onstantNode\x12\x63\n\x18request_data_source_node\x18\x04 \x01(\x0b\x32(.tecton_proto.args.RequestDataSourceNodeH\x00R\x15requestDataSourceNode\x12P\n\x11\x66\x65\x61ture_view_node\x18\x05 \x01(\x0b\x32\".tecton_proto.args.FeatureViewNodeH\x00R\x0f\x66\x65\x61tureViewNode\x12q\n\x1cmaterialization_context_node\x18\x07 \x01(\x0b\x32-.tecton_proto.args.MaterializationContextNodeH\x00R\x1amaterializationContextNodeB\x0b\n\tnode_typeJ\x04\x08\x06\x10\x07\"\x82\x01\n\x15RequestDataSourceNode\x12J\n\x0frequest_context\x18\x01 \x01(\x0b\x32!.tecton_proto.args.RequestContextR\x0erequestContext\x12\x1d\n\ninput_name\x18\x02 \x01(\tR\tinputName\"J\n\x0eRequestContext\x12\x38\n\x06schema\x18\x01 \x01(\x0b\x32 .tecton_proto.common.SparkSchemaR\x06schema\"\xc5\x01\n\x0f\x46\x65\x61tureViewNode\x12?\n\x0f\x66\x65\x61ture_view_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\rfeatureViewId\x12R\n\x0c\x66\x65\x61ture_view\x18\x02 \x01(\x0b\x32/.tecton_proto.args.FeatureServiceFeaturePackageR\x0b\x66\x65\x61tureView\x12\x1d\n\ninput_name\x18\x03 \x01(\tR\tinputName\"\x8c\x01\n\x12TransformationNode\x12\x44\n\x11transformation_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x10transformationId\x12\x30\n\x06inputs\x18\x02 \x03(\x0b\x32\x18.tecton_proto.args.InputR\x06inputs\"\x84\x01\n\x05Input\x12\x1d\n\targ_index\x18\x01 \x01(\x05H\x00R\x08\x61rgIndex\x12\x1b\n\x08\x61rg_name\x18\x02 \x01(\tH\x00R\x07\x61rgName\x12\x33\n\x04node\x18\x03 \x01(\x0b\x32\x1f.tecton_proto.args.PipelineNodeR\x04nodeB\n\n\x08\x61rg_type\"\xc4\x03\n\x0e\x44\x61taSourceNode\x12L\n\x16virtual_data_source_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x13virtualDataSourceId\x12\x33\n\x06window\x18\x02 \x01(\x0b\x32\x19.google.protobuf.DurationH\x00R\x06window\x12>\n\x1awindow_unbounded_preceding\x18\x04 \x01(\x08H\x00R\x18windowUnboundedPreceding\x12+\n\x10window_unbounded\x18\x06 \x01(\x08H\x00R\x0fwindowUnbounded\x12G\n\x11start_time_offset\x18\x07 \x01(\x0b\x32\x19.google.protobuf.DurationH\x00R\x0fstartTimeOffset\x12\x42\n\x0fschedule_offset\x18\x03 \x01(\x0b\x32\x19.google.protobuf.DurationR\x0escheduleOffset\x12\x1d\n\ninput_name\x18\x05 \x01(\tR\tinputNameB\x16\n\x14\x64\x61ta_window_override\"\xe0\x01\n\x0c\x43onstantNode\x12#\n\x0cstring_const\x18\x01 \x01(\tH\x00R\x0bstringConst\x12\x1d\n\tint_const\x18\x02 \x01(\tH\x00R\x08intConst\x12!\n\x0b\x66loat_const\x18\x03 \x01(\tH\x00R\nfloatConst\x12\x1f\n\nbool_const\x18\x04 \x01(\x08H\x00R\tboolConst\x12\x37\n\nnull_const\x18\x05 \x01(\x0b\x32\x16.google.protobuf.EmptyH\x00R\tnullConstB\x0f\n\rconstant_type\"\x1c\n\x1aMaterializationContextNode*\x9a\x01\n\x0cPipelineMode\x12\x19\n\x15PIPELINE_MODE_UNKNOWN\x10\x00\x12\x17\n\x13PIPELINE_MODE_SPARK\x10\x01\x12\x1b\n\x17PIPELINE_MODE_ON_DEMAND\x10\x02\x12\x1f\n\x1bPIPELINE_MODE_SNOWFLAKE_SQL\x10\x03\x12\x18\n\x14PIPELINE_MODE_ATHENA\x10\x04\x42\x13\n\x0f\x63om.tecton.argsP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.args.pipeline_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.argsP\001'
  _PIPELINEMODE._serialized_start=2326
  _PIPELINEMODE._serialized_end=2480
  _PIPELINE._serialized_start=227
  _PIPELINE._serialized_end=349
  _PIPELINENODE._serialized_start=352
  _PIPELINENODE._serialized_end=924
  _REQUESTDATASOURCENODE._serialized_start=927
  _REQUESTDATASOURCENODE._serialized_end=1057
  _REQUESTCONTEXT._serialized_start=1059
  _REQUESTCONTEXT._serialized_end=1133
  _FEATUREVIEWNODE._serialized_start=1136
  _FEATUREVIEWNODE._serialized_end=1333
  _TRANSFORMATIONNODE._serialized_start=1336
  _TRANSFORMATIONNODE._serialized_end=1476
  _INPUT._serialized_start=1479
  _INPUT._serialized_end=1611
  _DATASOURCENODE._serialized_start=1614
  _DATASOURCENODE._serialized_end=2066
  _CONSTANTNODE._serialized_start=2069
  _CONSTANTNODE._serialized_end=2293
  _MATERIALIZATIONCONTEXTNODE._serialized_start=2295
  _MATERIALIZATIONCONTEXTNODE._serialized_end=2323
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/api/featureservice/feature_service.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.api.featureservice import feature_service_request_pb2 as tecton__proto_dot_api_dot_featureservice_dot_feature__service__request__pb2
from google.api import annotations_pb2 as google_dot_api_dot_annotations__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from protoc_gen_swagger.options import annotations_pb2 as protoc__gen__swagger_dot_options_dot_annotations__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n5tecton_proto/api/featureservice/feature_service.proto\x12 tecton_proto.api.feature_service\x1a=tecton_proto/api/featureservice/feature_service_request.proto\x1a\x1cgoogle/api/annotations.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1cgoogle/protobuf/struct.proto\x1a,protoc-gen-swagger/options/annotations.proto\"\xbf\x02\n\x0bNameAndType\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12[\n\tdata_type\x18\x03 \x01(\x0b\x32>.tecton_proto.api.feature_service.FeatureServerComplexDataTypeR\x08\x64\x61taType\x12O\n\x04type\x18\x02 \x01(\x0e\x32\x37.tecton_proto.api.feature_service.FeatureServerDataTypeB\x02\x18\x01R\x04type:n\x92\x41k\n\r*\x0bNameAndType:Z{\"name\":\"string\",\"dataType\": \"#/definitions/FeatureServerComplexDataType\",\"type\":\"string\"}\"\xe8\x01\n\x11GetFeaturesResult\x12\x32\n\x08\x66\x65\x61tures\x18\x05 \x03(\x0b\x32\x16.google.protobuf.ValueR\x08\x66\x65\x61tures\x12\x33\n\tjoin_keys\x18\x06 \x03(\x0b\x32\x16.google.protobuf.ValueR\x08joinKeys:R\x92\x41O\n\x13*\x11GetFeaturesResult:8{\"result\":{\"features\":[\"object\"],\"joinKeys\":[\"object\"]}}J\x04\x08\x01\x10\x02J\x04\x08\x02\x10\x03J\x04\x08\x03\x10\x04J\x04\x08\x04\x10\x05\"\xb6\x05\n\x1c\x46\x65\x61tureServerComplexDataType\x12K\n\x04type\x18\x01 \x01(\x0e\x32\x37.tecton_proto.api.feature_service.FeatureServerDataTypeR\x04type\x12\x61\n\x0c\x65lement_type\x18\x02 \x01(\x0b\x32>.tecton_proto.api.feature_service.FeatureServerComplexDataTypeR\x0b\x65lementType\x12R\n\x06\x66ields\x18\x03 \x03(\x0b\x32:.tecton_proto.api.feature_service.FeatureServerStructFieldR\x06\x66ields:\x91\x03\x92\x41\x8d\x03\n\x1e*\x1c\x46\x65\x61tureServerComplexDataType:\xea\x02{\"type\":\"string\",\"elementType\":{\"type\":\"string\",\"elementType\":\"#/definitions/FeatureServerComplexDataType\",\"fields\":[{\"name\":\"string\",\"dataType\":\"#/definitions/FeatureServerComplexDataType\"}]},\"fields\":[{\"name\":\"string\",\"dataType\":{\"type\":\"string\",\"elementType\":\"#/definitions/FeatureServerComplexDataType\",\"fields\":[\"#/definitions/FeatureServerStructField\"]}}]}\"\xb8\x04\n\x18\x46\x65\x61tureServerStructField\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12[\n\tdata_type\x18\x02 \x01(\x0b\x32>.tecton_proto.api.feature_service.FeatureServerComplexDataTypeR\x08\x64\x61taType:\xaa\x03\x92\x41\xa6\x03\n\x1a*\x18\x46\x65\x61tureServerStructField:\x87\x03{\"name\":\"string\",\"dataType\":{\"type\":\"string\",\"elementType\":{\"type\":\"string\",\"elementType\":\"#/definitions/FeatureServerComplexDataType\",\"fields\":[{\"name\":\"string\",\"dataType\":\"#/definitions/FeatureServerComplexDataType\"}]},\"fields\":[{\"name\":\"string\",\"dataType\":{\"type\":\"string\",\"elementType\":\"#/definitions/FeatureServerComplexDataType\",\"fields\":[\"#/definitions/FeatureServerStructField\"]}}]}}\"\xf9\x05\n\x07SloInfo\x12!\n\x0cslo_eligible\x18\x01 \x01(\x08R\x0bsloEligible\x12\x35\n\x17slo_server_time_seconds\x18\x05 \x01(\x01R\x14sloServerTimeSeconds\x12t\n\x19slo_ineligibility_reasons\x18\x02 \x03(\x0e\x32\x38.tecton_proto.api.feature_service.SloIneligibilityReasonR\x17sloIneligibilityReasons\x12?\n\x1c\x64ynamodb_response_size_bytes\x18\x03 \x01(\x05R\x19\x64ynamodbResponseSizeBytes\x12.\n\x13server_time_seconds\x18\x04 \x01(\x01R\x11serverTimeSeconds\x12,\n\x12store_time_seconds\x18\x06 \x01(\x01R\x10storeTimeSeconds\x12*\n\x11store_max_latency\x18\x07 \x01(\x01R\x0fstoreMaxLatency\x12\x39\n\x19store_response_size_bytes\x18\x08 \x01(\x05R\x16storeResponseSizeBytes:\x97\x02\x92\x41\x93\x02\n\t*\x07SloInfo:\x85\x02{\"sloEligible\":\"boolean\",\"sloServerTimeSeconds\":\"number(double)\",\"sloIneligibilityReasons\":[\"string\"],\"dynamodbResponseSizeBytes\":\"integer(int32)\",\"serverTimeSeconds\":\"number(double)\",\"storeMaxLatency\":\"number(double)\",\"storeResponseSizeBytes\":\"integer(int32)\"}\"\xd6\x12\n\x08Metadata\x12K\n\x08\x66\x65\x61tures\x18\x01 \x03(\x0b\x32/.tecton_proto.api.feature_service.Metadata.ItemR\x08\x66\x65\x61tures\x12L\n\tjoin_keys\x18\x02 \x03(\x0b\x32/.tecton_proto.api.feature_service.Metadata.ItemR\x08joinKeys\x12\x44\n\x08slo_info\x18\x03 \x01(\x0b\x32).tecton_proto.api.feature_service.SloInfoR\x07sloInfo\x12\'\n\x0fpartial_results\x18\x04 \x01(\x08R\x0epartialResults\x1a\xba\x06\n\x04Item\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x41\n\x0e\x65\x66\x66\x65\x63tive_time\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\reffectiveTime\x12O\n\x04type\x18\x03 \x01(\x0e\x32\x37.tecton_proto.api.feature_service.FeatureServerDataTypeB\x02\x18\x01R\x04type\x12[\n\tdata_type\x18\x04 \x01(\x0b\x32>.tecton_proto.api.feature_service.FeatureServerComplexDataTypeR\x08\x64\x61taType\x12@\n\x06status\x18\x06 \x01(\x0e\x32(.tecton_proto.api.feature_service.StatusR\x06status:\xe4\x03\x92\x41\xe0\x03\n\x0e*\x0cMetadataItem:\xcd\x03{\"name\":\"string\",\"effectiveTime\":\"string(date-time)\",\"type\":\"string\",\"dataType\":{\"type\":\"string\",\"elementType\":{\"type\":\"string\",\"elementType\":\"#/definitions/FeatureServerComplexDataType\",\"fields\":[{\"name\":\"string\",\"dataType\":\"#/definitions/FeatureServerComplexDataType\"}]},\"fields\":[{\"name\":\"string\",\"dataType\":{\"type\":\"string\",\"elementType\":\"#/definitions/FeatureServerComplexDataType\",\"fields\":[\"#/definitions/FeatureServerStructField\"]}}]},\"status\":\"string\"}J\x04\x08\x05\x10\x06:\x82\n\x92\x41\xfe\t\n\n*\x08Metadata:\xef\t{\"metadata\":{\"features\":[{\"name\":\"string\",\"effectiveTime\":\"string(date-time)\",\"type\":\"string\",\"dataType\":{\"type\":\"string\",\"elementType\":{\"type\":\"string\",\"elementType\":\"#/definitions/FeatureServerComplexDataType\",\"fields\":[{\"name\":\"string\",\"dataType\":\"#/definitions/FeatureServerComplexDataType\"}]},\"fields\":[{\"name\":\"string\",\"dataType\":{\"type\":\"string\",\"elementType\":\"#/definitions/FeatureServerComplexDataType\",\"fields\":[\"#/definitions/FeatureServerStructField\"]}}]},\"status\":\"string\"}],\"joinKeys\":[{\"name\":\"string\",\"effectiveTime\":\"string(date-time)\",\"type\":\"string\",\"dataType\":{\"type\":\"string\",\"elementType\":{\"type\":\"string\",\"elementType\":\"#/definitions/FeatureServerComplexDataType\",\"fields\":[{\"name\":\"string\",\"dataType\":\"#/definitions/FeatureServerComplexDataType\"}]},\"fields\":[{\"name\":\"string\",\"dataType\":{\"type\":\"string\",\"elementType\":\"#/definitions/FeatureServerComplexDataType\",\"fields\":[\"#/definitions/FeatureServerStructField\"]}}]},\"status\":\"string\"}],\"sloInfo\":{\"sloEligible\":\"boolean\",\"sloServerTimeSeconds\":\"number(double)\",\"sloIneligibilityReasons\":[\"string\"],\"dynamodbResponseSizeBytes\":\"integer(int32)\",\"serverTimeSeconds\":\"number(double)\",\"storeMaxLatency\":\"number(double)\",\"storeResponseSizeBytes\":\"integer(int32)\"},\"partialResults\":\"boolean\"}}\"\xaa\x01\n\x13GetFeaturesResponse\x12K\n\x06result\x18\x02 \x01(\x0b\x32\x33.tecton_proto.api.feature_service.GetFeaturesResultR\x06result\x12\x46\n\x08metadata\x18\x03 \x01(\x0b\x32*.tecton_proto.api.feature_service.MetadataR\x08metadata\"\xba\x01\n\x18GetFeaturesBatchResponse\x12K\n\x06result\x18\x02 \x03(\x0b\x32\x33.tecton_proto.api.feature_service.GetFeaturesResultR\x06result\x12K\n\x08metadata\x18\x04 \x01(\x0b\x32/.tecton_proto.api.feature_service.BatchMetadataR\x08metadataJ\x04\x08\x03\x10\x04\"\xba\x02\n\x0c\x42\x61tchSloInfo\x12!\n\x0cslo_eligible\x18\x01 \x01(\x08R\x0bsloEligible\x12\x35\n\x17slo_server_time_seconds\x18\x02 \x01(\x01R\x14sloServerTimeSeconds\x12t\n\x19slo_ineligibility_reasons\x18\x03 \x03(\x0e\x32\x38.tecton_proto.api.feature_service.SloIneligibilityReasonR\x17sloIneligibilityReasons\x12.\n\x13server_time_seconds\x18\x04 \x01(\x01R\x11serverTimeSeconds\x12*\n\x11store_max_latency\x18\x05 \x01(\x01R\x0fstoreMaxLatency\"\xd5\x04\n\rBatchMetadata\x12P\n\x08\x66\x65\x61tures\x18\x01 \x03(\x0b\x32\x34.tecton_proto.api.feature_service.BatchMetadata.ItemR\x08\x66\x65\x61tures\x12Q\n\tjoin_keys\x18\x02 \x03(\x0b\x32\x34.tecton_proto.api.feature_service.BatchMetadata.ItemR\x08joinKeys\x12\x44\n\x08slo_info\x18\x03 \x03(\x0b\x32).tecton_proto.api.feature_service.SloInfoR\x07sloInfo\x12T\n\x0e\x62\x61tch_slo_info\x18\x04 \x01(\x0b\x32..tecton_proto.api.feature_service.BatchSloInfoR\x0c\x62\x61tchSloInfo\x1a\x82\x02\n\x04Item\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x41\n\x0e\x65\x66\x66\x65\x63tive_time\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\reffectiveTime\x12[\n\tdata_type\x18\x03 \x01(\x0b\x32>.tecton_proto.api.feature_service.FeatureServerComplexDataTypeR\x08\x64\x61taType\x12@\n\x06status\x18\x05 \x03(\x0e\x32(.tecton_proto.api.feature_service.StatusR\x06statusJ\x04\x08\x04\x10\x05\"i\n\x16ServiceMetadataRequest\x12O\n\x06params\x18\x01 \x01(\x0b\x32\x37.tecton_proto.api.feature_service.FeatureServiceLocatorR\x06params\"\xfa\x05\n\x17ServiceMetadataResponse\x12\x66\n\x14\x66\x65\x61ture_service_type\x18\x01 \x01(\x0e\x32\x34.tecton_proto.api.feature_service.FeatureServiceTypeR\x12\x66\x65\x61tureServiceType\x12U\n\x0finput_join_keys\x18\x02 \x03(\x0b\x32-.tecton_proto.api.feature_service.NameAndTypeR\rinputJoinKeys\x12j\n\x1ainput_request_context_keys\x18\x03 \x03(\x0b\x32-.tecton_proto.api.feature_service.NameAndTypeR\x17inputRequestContextKeys\x12W\n\x10output_join_keys\x18\x04 \x03(\x0b\x32-.tecton_proto.api.feature_service.NameAndTypeR\x0eoutputJoinKeys\x12T\n\x0e\x66\x65\x61ture_values\x18\x05 \x03(\x0b\x32-.tecton_proto.api.feature_service.NameAndTypeR\rfeatureValues:\x84\x02\x92\x41\x80\x02\n\x19*\x17ServiceMetadataResponse:\xe2\x01{\"featureServiceType\":\"string\",\"inputJoinKeys\":[\"#/definitions/NameAndType\"],\"inputRequestContextKeys\":[\"#/definitions/NameAndType\"],\"outputJoinKeys\":[\"#/definitions/NameAndType\"],\"featureValues\":[\"#/definitions/NameAndType\"]}\"g\n\x14QueryFeaturesRequest\x12O\n\x06params\x18\x01 \x01(\x0b\x32\x37.tecton_proto.api.feature_service.GetFeaturesParametersR\x06params\"\xae\x01\n\x15QueryFeaturesResponse\x12M\n\x07results\x18\x01 \x03(\x0b\x32\x33.tecton_proto.api.feature_service.GetFeaturesResultR\x07results\x12\x46\n\x08metadata\x18\x02 \x01(\x0b\x32*.tecton_proto.api.feature_service.MetadataR\x08metadata\"\xb7\x03\n\x15\x46\x65\x61tureServiceLocator\x12.\n\x12\x66\x65\x61ture_service_id\x18\x01 \x01(\tH\x00R\x10\x66\x65\x61tureServiceId\x12\x32\n\x14\x66\x65\x61ture_service_name\x18\x02 \x01(\tH\x00R\x12\x66\x65\x61tureServiceName\x12.\n\x12\x66\x65\x61ture_package_id\x18\x04 \x01(\tH\x00R\x10\x66\x65\x61turePackageId\x12\x32\n\x14\x66\x65\x61ture_package_name\x18\x05 \x01(\tH\x00R\x12\x66\x65\x61turePackageName\x12(\n\x0f\x66\x65\x61ture_view_id\x18\x06 \x01(\tH\x00R\rfeatureViewId\x12,\n\x11\x66\x65\x61ture_view_name\x18\x07 \x01(\tH\x00R\x0f\x66\x65\x61tureViewName\x12%\n\x0eworkspace_name\x18\x03 \x01(\tR\rworkspaceName:L\x92\x41I:G{\"featureServiceName\": \"ctrPredictionService\", \"workspaceName\": \"prod\"}B\t\n\x07locator\"\x90\x01\n\x1dGetFeatureServiceStateRequest\x12o\n\x17\x66\x65\x61ture_service_locator\x18\x01 \x01(\x0b\x32\x37.tecton_proto.api.feature_service.FeatureServiceLocatorR\x15\x66\x65\x61tureServiceLocator\"&\n\x1eGetFeatureServiceStateResponseJ\x04\x08\x01\x10\x02*/\n\x12\x46\x65\x61tureServiceType\x12\x0b\n\x07\x44\x45\x46\x41ULT\x10\x00\x12\x0c\n\x08WILDCARD\x10\x01*\x90\x01\n\x15\x46\x65\x61tureServerDataType\x12\x10\n\x0cmissing_type\x10\x00\x12\x0b\n\x07\x62oolean\x10\x01\x12\x0b\n\x07\x66loat64\x10\x02\x12\t\n\x05int64\x10\x03\x12\n\n\x06string\x10\x04\x12\x10\n\x0cstring_array\x10\x05\x12\t\n\x05\x61rray\x10\x06\x12\x0b\n\x07\x66loat32\x10\x07\x12\n\n\x06struct\x10\x08*;\n\x06Status\x12\x0b\n\x07PRESENT\x10\x00\x12\x10\n\x0cMISSING_DATA\x10\x01\x12\x12\n\x0eUNKNOWN_STATUS\x10\x02*\x9a\x01\n\x16SloIneligibilityReason\x12\x0b\n\x07UNKNOWN\x10\x00\x12)\n%DYNAMODB_RESPONSE_SIZE_LIMIT_EXCEEDED\x10\x01\x12&\n\"REDIS_RESPONSE_SIZE_LIMIT_EXCEEDED\x10\x02\x12 \n\x1cREDIS_LATENCY_LIMIT_EXCEEDED\x10\x03\x32\xf1-\n\x0e\x46\x65\x61tureService\x12\x84\x12\n\x0bGetFeatures\x12\x34.tecton_proto.api.feature_service.GetFeaturesRequest\x1a\x35.tecton_proto.api.feature_service.GetFeaturesResponse\"\x87\x11\x82\xd3\xe4\x93\x02%\" /v1/feature-service/get-features:\x01*\x92\x41\xd8\x10\n\x0bGetFeatures\x1a\x46Get Features API to retrieve feature values from Tecton\'s online storeJ?\n\x03\x32\x30\x30\x12\x38\n\x02OK\"2\n\x10\x61pplication/json\x12\x1e{\"result\": {\"features\": [15]}}J\x97\x03\n\x03\x34\x30\x30\x12\x8f\x03\n8Bad Request : Invalid argument (invalid request payload)\"\xd2\x02\n\x10\x61pplication/json\x12\xbd\x02{\"join key error\": \"Unknown join key: content_id_withTypo, expected one of: [content_id]\", \"missing key error\": \"Unknown join key: Expected key not found: join_key\", \"incorrect type error\": \"Unable to parse \'type\' as an type for GetFeatures.\", \"dynamo db error\": \"sequential DynamoDB query limit exceeded\", \"code\": 3}J\x8a\x03\n\x03\x34\x30\x31\x12\x82\x03\nTNot Authenticated. Newly created credentials may take up to 60 seconds to be usable.\"\xa9\x02\n\x10\x61pplication/json\x12\x94\x02{\"error\": \"invalid \'Tecton-key\' authorization header. Note that newly created credentials may take up to 60 seconds to be usable.\", \"message\": \"invalid \'Tecton-key\' authorization header. Note that newly created credentials may take up to 60 seconds to be usable.\", \"code\": 16}J\xcb\x01\n\x03\x34\x30\x33\x12\xc3\x01\nhThe caller is not authorized for the request. Access control changes may take up to 60 seconds to apply.\"W\n\x10\x61pplication/json\x12\x43{\"error\": \"Not Authorized\", \"message\": \"Not Authorized\", \"code\": 7}J\xd6\x03\n\x03\x34\x30\x34\x12\xce\x03\n\tNot Found\"\xc0\x03\n\x10\x61pplication/json\x12\xab\x03{\"feature service id error\": \"Unknown feature_service_id: id\", \"feature service workspace error\": \"Unable to query FeatureService `fs` for workspace `ws`. Newly created feature services may take up to 60 seconds to query. Also, ensure that the workspace is a live workspace.\", \"feature package id error\": \"Unknown feature_package_id: id\", \"feature view name error\": \"Unknown feature_view_name: fv for workspace: ws\", \"code\": 5}J\xe5\x01\n\x03\x34\x32\x39\x12\xdd\x01\n\x12Resource Exhausted\"\xc6\x01\n\x10\x61pplication/json\x12\xb1\x01{\"error\": \"GetFeatures exceeded the concurrent request limit, please retry later\", \"message\": \"GetFeatures exceeded the concurrent request limit, please retry later\", \"code\": 8}J\x9f\x01\n\x03\x35\x30\x33\x12\x97\x01\n\x1fService Temporarily Unavailable\"t\n\x10\x61pplication/json\x12`{\"error\": \"individual request 5xx error\", \"message\": \"individual request 5xx error\", \"code\": 14}Jh\n\x03\x35\x30\x34\x12\x61\n\x11Gateway Timed Out\"L\n\x10\x61pplication/json\x12\x38{\"error\": \"Timedout\", \"message\": \"Timed out\", \"code\": 4}\x12\xd3\t\n\x10GetFeaturesBatch\x12\x39.tecton_proto.api.feature_service.GetFeaturesBatchRequest\x1a:.tecton_proto.api.feature_service.GetFeaturesBatchResponse\"\xc7\x08\x82\xd3\xe4\x93\x02+\"&/v1/feature-service/get-features-batch:\x01*\x92\x41\x92\x08\n\x10GetFeaturesBatch\x1aUBatch get Features API to retrieve feature values in batch from Tecton\'s online storeJ^\n\x03\x32\x30\x30\x12W\n\x02OK\"Q\n\x10\x61pplication/json\x12={\"result\": {\"features\": [[15, 10, 5, 100],[10, 20, 1, 200]]}}J\xae\x01\n\x03\x34\x30\x34\x12\xa6\x01\n\tNot Found\"\x98\x01\n\x10\x61pplication/json\x12\x83\x01{\"error\": \"Unknown feature_service_id: my_feature_service\", \"message\": \"Unknown feature_service_id: my_feature_service\", \"code\": 5}J\x9b\x01\n\x03\x34\x30\x30\x12\x93\x01\n\x0b\x42\x61\x64 Request\"\x83\x01\n\x10\x61pplication/json\x12o{\"error\": \"Batch size 27 exceeds maximum of 25.\", \"message\": \"Batch size 27 exceeds maximum of 25.\", \"code\": 3}J\xea\x01\n\x03\x34\x32\x39\x12\xe2\x01\n\x12Resource Exhausted\"\xcb\x01\n\x10\x61pplication/json\x12\xb6\x01{\"error\": \"GetFeaturesBatch exceeded the concurrent request limit, please retry later\", \"message\": \"GetFeatures exceeded the concurrent request limit, please retry later\", \"code\": 8}J\x9f\x01\n\x03\x35\x30\x33\x12\x97\x01\n\x1fService Temporarily Unavailable\"t\n\x10\x61pplication/json\x12`{\"error\": \"individual request 5xx error\", \"message\": \"individual request 5xx error\", \"code\": 14}Ji\n\x03\x35\x30\x34\x12\x62\n\x11Gateway Timed Out\"M\n\x10\x61pplication/json\x12\x39{\"error\": \"Timed out\", \"message\": \"Timed out\", \"code\": 4}\x12\xaf\x01\n\rQueryFeatures\x12\x36.tecton_proto.api.feature_service.QueryFeaturesRequest\x1a\x37.tecton_proto.api.feature_service.QueryFeaturesResponse\"-\x82\xd3\xe4\x93\x02\'\"\"/v1/feature-service/query-features:\x01*\x12\xd5\x01\n\x16GetFeatureServiceState\x12?.tecton_proto.api.feature_service.GetFeatureServiceStateRequest\x1a@.tecton_proto.api.feature_service.GetFeatureServiceStateResponse\"8\x82\xd3\xe4\x93\x02\x32\"-/v1/feature-service/get-feature-service-state:\x01*\x12\xf7\x0e\n\x0fServiceMetadata\x12\x38.tecton_proto.api.feature_service.ServiceMetadataRequest\x1a\x39.tecton_proto.api.feature_service.ServiceMetadataResponse\"\xee\r\x82\xd3\xe4\x93\x02!\"\x1c/v1/feature-service/metadata:\x01*\x92\x41\xc3\r\n\x08Metadata\x1a\x32Get Metadata about the feature service from TectonJ\xd8\x02\n\x03\x32\x30\x30\x12\xd0\x02\n\x02OK\"\xc9\x02\n\x10\x61pplication/json\x12\xb4\x02{\"featureServiceType\": \"DEFAULT\", \"inputJoinKeys\": [{\"name\": \"adId\", \"type\": \"string\"}],\"featureValues\":[{\"name\": \"ad_ground_truth_ctr_performance_7_days\", \"type\": \"float64\"}, {\"name\": \"user_partner_impression_count_7_days\", \"type\": \"float64\"}, {\"name\": \"user_total_ad_frequency_counts\", \"type\": \"int64\"} ] }J\x8a\x03\n\x03\x34\x30\x31\x12\x82\x03\nTNot Authenticated. Newly created credentials may take up to 60 seconds to be usable.\"\xa9\x02\n\x10\x61pplication/json\x12\x94\x02{\"error\": \"invalid \'Tecton-key\' authorization header. Note that newly created credentials may take up to 60 seconds to be usable.\", \"message\": \"invalid \'Tecton-key\' authorization header. Note that newly created credentials may take up to 60 seconds to be usable.\", \"code\": 16}J\xcb\x01\n\x03\x34\x30\x33\x12\xc3\x01\nhThe caller is not authorized for the request. Access control changes may take up to 60 seconds to apply.\"W\n\x10\x61pplication/json\x12\x43{\"error\": \"Not Authorized\", \"message\": \"Not Authorized\", \"code\": 7}J\xda\x01\n\x03\x34\x30\x34\x12\xd2\x01\n\tNot Found\"\xc4\x01\n\x10\x61pplication/json\x12\xaf\x01{\"error\": \"Unknown feature_service_name: my_feature_service for workspace: prod\", \"message\": \"Unknown feature_service_name: my_feature_service for workspace: prod\", \"code\": 5}J\xe2\x01\n\x03\x34\x32\x39\x12\xda\x01\n\x12Resource Exhausted\"\xc3\x01\n\x10\x61pplication/json\x12\xae\x01{\"error\": \"Metadata exceeded the concurrent request limit, please retry later\", \"message\": \"GetFeatures exceeded the concurrent request limit, please retry later\", \"code\": 8}J\x9f\x01\n\x03\x35\x30\x33\x12\x97\x01\n\x1fService Temporarily Unavailable\"t\n\x10\x61pplication/json\x12`{\"error\": \"individual request 5xx error\", \"message\": \"individual request 5xx error\", \"code\": 14}Ji\n\x03\x35\x30\x34\x12\x62\n\x11Gateway Timed Out\"M\n\x10\x61pplication/json\x12\x39{\"error\": \"Timed out\", \"message\": \"Timed out\", \"code\": 4}B\xd6\nZ4github.com/tecton-ai/tecton_proto/api/featureservice\x92\x41\x9c\n\x12\xf9\x02\n\x13\x46\x65\x61ture Service API\x12\xc0\x02Tecton provides a low-latency feature server that exposes HTTP endpoints to retrieve feature values and metadata from the online store. These endpoints are typically used during model predictions. The feature servers retrieve data from the online store and perform any additional aggregations and filtering as necessary.\"\x0b\n\tTecton AI*\r\n\x0bProprietary2\x03\x31.0*\x01\x02\x32\x10\x61pplication/json:\x10\x61pplication/jsonR\x8a\x03\n\x03\x34\x30\x31\x12\x82\x03\nTNot Authenticated. Newly created credentials may take up to 60 seconds to be usable.\"\xa9\x02\n\x10\x61pplication/json\x12\x94\x02{\"error\": \"invalid \'Tecton-key\' authorization header. Note that newly created credentials may take up to 60 seconds to be usable.\", \"code\": 16, \"message\": \"invalid \'Tecton-key\' authorization header. Note that newly created credentials may take up to 60 seconds to be usable.\"}R\xd6\x01\n\x03\x34\x30\x33\x12\xce\x01\nsForbidden. The caller is not authorized for the request. Access control changes may take up to 60 seconds to apply.\"W\n\x10\x61pplication/json\x12\x43{\"error\": \"Not Authorized\", \"code\": 7, \"message\": \"Not Authorized\"}Z\xfa\x01\n\xf7\x01\n\x0eTecton API Key\x12\xe4\x01\x08\x02\x12\xce\x01Use your Tecton API Key (prefixed by `Tecton-key`) as the `Authorization` request header field. For example, for the API Key `secretapikey123`, use `Tecton-key secretapikey123` as your authorization header.\x1a\rAuthorization \x02\x62\x14\n\x12\n\x0eTecton API Key\x12\x00')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.api.featureservice.feature_service_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z4github.com/tecton-ai/tecton_proto/api/featureservice\222A\234\n\022\371\002\n\023Feature Service API\022\300\002Tecton provides a low-latency feature server that exposes HTTP endpoints to retrieve feature values and metadata from the online store. These endpoints are typically used during model predictions. The feature servers retrieve data from the online store and perform any additional aggregations and filtering as necessary.\"\013\n\tTecton AI*\r\n\013Proprietary2\0031.0*\001\0022\020application/json:\020application/jsonR\212\003\n\003401\022\202\003\nTNot Authenticated. Newly created credentials may take up to 60 seconds to be usable.\"\251\002\n\020application/json\022\224\002{\"error\": \"invalid \'Tecton-key\' authorization header. Note that newly created credentials may take up to 60 seconds to be usable.\", \"code\": 16, \"message\": \"invalid \'Tecton-key\' authorization header. Note that newly created credentials may take up to 60 seconds to be usable.\"}R\326\001\n\003403\022\316\001\nsForbidden. The caller is not authorized for the request. Access control changes may take up to 60 seconds to apply.\"W\n\020application/json\022C{\"error\": \"Not Authorized\", \"code\": 7, \"message\": \"Not Authorized\"}Z\372\001\n\367\001\n\016Tecton API Key\022\344\001\010\002\022\316\001Use your Tecton API Key (prefixed by `Tecton-key`) as the `Authorization` request header field. For example, for the API Key `secretapikey123`, use `Tecton-key secretapikey123` as your authorization header.\032\rAuthorization \002b\024\n\022\n\016Tecton API Key\022\000'
  _NAMEANDTYPE.fields_by_name['type']._options = None
  _NAMEANDTYPE.fields_by_name['type']._serialized_options = b'\030\001'
  _NAMEANDTYPE._options = None
  _NAMEANDTYPE._serialized_options = b'\222Ak\n\r*\013NameAndType:Z{\"name\":\"string\",\"dataType\": \"#/definitions/FeatureServerComplexDataType\",\"type\":\"string\"}'
  _GETFEATURESRESULT._options = None
  _GETFEATURESRESULT._serialized_options = b'\222AO\n\023*\021GetFeaturesResult:8{\"result\":{\"features\":[\"object\"],\"joinKeys\":[\"object\"]}}'
  _FEATURESERVERCOMPLEXDATATYPE._options = None
  _FEATURESERVERCOMPLEXDATATYPE._serialized_options = b'\222A\215\003\n\036*\034FeatureServerComplexDataType:\352\002{\"type\":\"string\",\"elementType\":{\"type\":\"string\",\"elementType\":\"#/definitions/FeatureServerComplexDataType\",\"fields\":[{\"name\":\"string\",\"dataType\":\"#/definitions/FeatureServerComplexDataType\"}]},\"fields\":[{\"name\":\"string\",\"dataType\":{\"type\":\"string\",\"elementType\":\"#/definitions/FeatureServerComplexDataType\",\"fields\":[\"#/definitions/FeatureServerStructField\"]}}]}'
  _FEATURESERVERSTRUCTFIELD._options = None
  _FEATURESERVERSTRUCTFIELD._serialized_options = b'\222A\246\003\n\032*\030FeatureServerStructField:\207\003{\"name\":\"string\",\"dataType\":{\"type\":\"string\",\"elementType\":{\"type\":\"string\",\"elementType\":\"#/definitions/FeatureServerComplexDataType\",\"fields\":[{\"name\":\"string\",\"dataType\":\"#/definitions/FeatureServerComplexDataType\"}]},\"fields\":[{\"name\":\"string\",\"dataType\":{\"type\":\"string\",\"elementType\":\"#/definitions/FeatureServerComplexDataType\",\"fields\":[\"#/definitions/FeatureServerStructField\"]}}]}}'
  _SLOINFO._options = None
  _SLOINFO._serialized_options = b'\222A\223\002\n\t*\007SloInfo:\205\002{\"sloEligible\":\"boolean\",\"sloServerTimeSeconds\":\"number(double)\",\"sloIneligibilityReasons\":[\"string\"],\"dynamodbResponseSizeBytes\":\"integer(int32)\",\"serverTimeSeconds\":\"number(double)\",\"storeMaxLatency\":\"number(double)\",\"storeResponseSizeBytes\":\"integer(int32)\"}'
  _METADATA_ITEM.fields_by_name['type']._options = None
  _METADATA_ITEM.fields_by_name['type']._serialized_options = b'\030\001'
  _METADATA_ITEM._options = None
  _METADATA_ITEM._serialized_options = b'\222A\340\003\n\016*\014MetadataItem:\315\003{\"name\":\"string\",\"effectiveTime\":\"string(date-time)\",\"type\":\"string\",\"dataType\":{\"type\":\"string\",\"elementType\":{\"type\":\"string\",\"elementType\":\"#/definitions/FeatureServerComplexDataType\",\"fields\":[{\"name\":\"string\",\"dataType\":\"#/definitions/FeatureServerComplexDataType\"}]},\"fields\":[{\"name\":\"string\",\"dataType\":{\"type\":\"string\",\"elementType\":\"#/definitions/FeatureServerComplexDataType\",\"fields\":[\"#/definitions/FeatureServerStructField\"]}}]},\"status\":\"string\"}'
  _METADATA._options = None
  _METADATA._serialized_options = b'\222A\376\t\n\n*\010Metadata:\357\t{\"metadata\":{\"features\":[{\"name\":\"string\",\"effectiveTime\":\"string(date-time)\",\"type\":\"string\",\"dataType\":{\"type\":\"string\",\"elementType\":{\"type\":\"string\",\"elementType\":\"#/definitions/FeatureServerComplexDataType\",\"fields\":[{\"name\":\"string\",\"dataType\":\"#/definitions/FeatureServerComplexDataType\"}]},\"fields\":[{\"name\":\"string\",\"dataType\":{\"type\":\"string\",\"elementType\":\"#/definitions/FeatureServerComplexDataType\",\"fields\":[\"#/definitions/FeatureServerStructField\"]}}]},\"status\":\"string\"}],\"joinKeys\":[{\"name\":\"string\",\"effectiveTime\":\"string(date-time)\",\"type\":\"string\",\"dataType\":{\"type\":\"string\",\"elementType\":{\"type\":\"string\",\"elementType\":\"#/definitions/FeatureServerComplexDataType\",\"fields\":[{\"name\":\"string\",\"dataType\":\"#/definitions/FeatureServerComplexDataType\"}]},\"fields\":[{\"name\":\"string\",\"dataType\":{\"type\":\"string\",\"elementType\":\"#/definitions/FeatureServerComplexDataType\",\"fields\":[\"#/definitions/FeatureServerStructField\"]}}]},\"status\":\"string\"}],\"sloInfo\":{\"sloEligible\":\"boolean\",\"sloServerTimeSeconds\":\"number(double)\",\"sloIneligibilityReasons\":[\"string\"],\"dynamodbResponseSizeBytes\":\"integer(int32)\",\"serverTimeSeconds\":\"number(double)\",\"storeMaxLatency\":\"number(double)\",\"storeResponseSizeBytes\":\"integer(int32)\"},\"partialResults\":\"boolean\"}}'
  _SERVICEMETADATARESPONSE._options = None
  _SERVICEMETADATARESPONSE._serialized_options = b'\222A\200\002\n\031*\027ServiceMetadataResponse:\342\001{\"featureServiceType\":\"string\",\"inputJoinKeys\":[\"#/definitions/NameAndType\"],\"inputRequestContextKeys\":[\"#/definitions/NameAndType\"],\"outputJoinKeys\":[\"#/definitions/NameAndType\"],\"featureValues\":[\"#/definitions/NameAndType\"]}'
  _FEATURESERVICELOCATOR._options = None
  _FEATURESERVICELOCATOR._serialized_options = b'\222AI:G{\"featureServiceName\": \"ctrPredictionService\", \"workspaceName\": \"prod\"}'
  _FEATURESERVICE.methods_by_name['GetFeatures']._options = None
  _FEATURESERVICE.methods_by_name['GetFeatures']._serialized_options = b'\202\323\344\223\002%\" /v1/feature-service/get-features:\001*\222A\330\020\n\013GetFeatures\032FGet Features API to retrieve feature values from Tecton\'s online storeJ?\n\003200\0228\n\002OK\"2\n\020application/json\022\036{\"result\": {\"features\": [15]}}J\227\003\n\003400\022\217\003\n8Bad Request : Invalid argument (invalid request payload)\"\322\002\n\020application/json\022\275\002{\"join key error\": \"Unknown join key: content_id_withTypo, expected one of: [content_id]\", \"missing key error\": \"Unknown join key: Expected key not found: join_key\", \"incorrect type error\": \"Unable to parse \'type\' as an type for GetFeatures.\", \"dynamo db error\": \"sequential DynamoDB query limit exceeded\", \"code\": 3}J\212\003\n\003401\022\202\003\nTNot Authenticated. Newly created credentials may take up to 60 seconds to be usable.\"\251\002\n\020application/json\022\224\002{\"error\": \"invalid \'Tecton-key\' authorization header. Note that newly created credentials may take up to 60 seconds to be usable.\", \"message\": \"invalid \'Tecton-key\' authorization header. Note that newly created credentials may take up to 60 seconds to be usable.\", \"code\": 16}J\313\001\n\003403\022\303\001\nhThe caller is not authorized for the request. Access control changes may take up to 60 seconds to apply.\"W\n\020application/json\022C{\"error\": \"Not Authorized\", \"message\": \"Not Authorized\", \"code\": 7}J\326\003\n\003404\022\316\003\n\tNot Found\"\300\003\n\020application/json\022\253\003{\"feature service id error\": \"Unknown feature_service_id: id\", \"feature service workspace error\": \"Unable to query FeatureService `fs` for workspace `ws`. Newly created feature services may take up to 60 seconds to query. Also, ensure that the workspace is a live workspace.\", \"feature package id error\": \"Unknown feature_package_id: id\", \"feature view name error\": \"Unknown feature_view_name: fv for workspace: ws\", \"code\": 5}J\345\001\n\003429\022\335\001\n\022Resource Exhausted\"\306\001\n\020application/json\022\261\001{\"error\": \"GetFeatures exceeded the concurrent request limit, please retry later\", \"message\": \"GetFeatures exceeded the concurrent request limit, please retry later\", \"code\": 8}J\237\001\n\003503\022\227\001\n\037Service Temporarily Unavailable\"t\n\020application/json\022`{\"error\": \"individual request 5xx error\", \"message\": \"individual request 5xx error\", \"code\": 14}Jh\n\003504\022a\n\021Gateway Timed Out\"L\n\020application/json\0228{\"error\": \"Timedout\", \"message\": \"Timed out\", \"code\": 4}'
  _FEATURESERVICE.methods_by_name['GetFeaturesBatch']._options = None
  _FEATURESERVICE.methods_by_name['GetFeaturesBatch']._serialized_options = b'\202\323\344\223\002+\"&/v1/feature-service/get-features-batch:\001*\222A\222\010\n\020GetFeaturesBatch\032UBatch get Features API to retrieve feature values in batch from Tecton\'s online storeJ^\n\003200\022W\n\002OK\"Q\n\020application/json\022={\"result\": {\"features\": [[15, 10, 5, 100],[10, 20, 1, 200]]}}J\256\001\n\003404\022\246\001\n\tNot Found\"\230\001\n\020application/json\022\203\001{\"error\": \"Unknown feature_service_id: my_feature_service\", \"message\": \"Unknown feature_service_id: my_feature_service\", \"code\": 5}J\233\001\n\003400\022\223\001\n\013Bad Request\"\203\001\n\020application/json\022o{\"error\": \"Batch size 27 exceeds maximum of 25.\", \"message\": \"Batch size 27 exceeds maximum of 25.\", \"code\": 3}J\352\001\n\003429\022\342\001\n\022Resource Exhausted\"\313\001\n\020application/json\022\266\001{\"error\": \"GetFeaturesBatch exceeded the concurrent request limit, please retry later\", \"message\": \"GetFeatures exceeded the concurrent request limit, please retry later\", \"code\": 8}J\237\001\n\003503\022\227\001\n\037Service Temporarily Unavailable\"t\n\020application/json\022`{\"error\": \"individual request 5xx error\", \"message\": \"individual request 5xx error\", \"code\": 14}Ji\n\003504\022b\n\021Gateway Timed Out\"M\n\020application/json\0229{\"error\": \"Timed out\", \"message\": \"Timed out\", \"code\": 4}'
  _FEATURESERVICE.methods_by_name['QueryFeatures']._options = None
  _FEATURESERVICE.methods_by_name['QueryFeatures']._serialized_options = b'\202\323\344\223\002\'\"\"/v1/feature-service/query-features:\001*'
  _FEATURESERVICE.methods_by_name['GetFeatureServiceState']._options = None
  _FEATURESERVICE.methods_by_name['GetFeatureServiceState']._serialized_options = b'\202\323\344\223\0022\"-/v1/feature-service/get-feature-service-state:\001*'
  _FEATURESERVICE.methods_by_name['ServiceMetadata']._options = None
  _FEATURESERVICE.methods_by_name['ServiceMetadata']._serialized_options = b'\202\323\344\223\002!\"\034/v1/feature-service/metadata:\001*\222A\303\r\n\010Metadata\0322Get Metadata about the feature service from TectonJ\330\002\n\003200\022\320\002\n\002OK\"\311\002\n\020application/json\022\264\002{\"featureServiceType\": \"DEFAULT\", \"inputJoinKeys\": [{\"name\": \"adId\", \"type\": \"string\"}],\"featureValues\":[{\"name\": \"ad_ground_truth_ctr_performance_7_days\", \"type\": \"float64\"}, {\"name\": \"user_partner_impression_count_7_days\", \"type\": \"float64\"}, {\"name\": \"user_total_ad_frequency_counts\", \"type\": \"int64\"} ] }J\212\003\n\003401\022\202\003\nTNot Authenticated. Newly created credentials may take up to 60 seconds to be usable.\"\251\002\n\020application/json\022\224\002{\"error\": \"invalid \'Tecton-key\' authorization header. Note that newly created credentials may take up to 60 seconds to be usable.\", \"message\": \"invalid \'Tecton-key\' authorization header. Note that newly created credentials may take up to 60 seconds to be usable.\", \"code\": 16}J\313\001\n\003403\022\303\001\nhThe caller is not authorized for the request. Access control changes may take up to 60 seconds to apply.\"W\n\020application/json\022C{\"error\": \"Not Authorized\", \"message\": \"Not Authorized\", \"code\": 7}J\332\001\n\003404\022\322\001\n\tNot Found\"\304\001\n\020application/json\022\257\001{\"error\": \"Unknown feature_service_name: my_feature_service for workspace: prod\", \"message\": \"Unknown feature_service_name: my_feature_service for workspace: prod\", \"code\": 5}J\342\001\n\003429\022\332\001\n\022Resource Exhausted\"\303\001\n\020application/json\022\256\001{\"error\": \"Metadata exceeded the concurrent request limit, please retry later\", \"message\": \"GetFeatures exceeded the concurrent request limit, please retry later\", \"code\": 8}J\237\001\n\003503\022\227\001\n\037Service Temporarily Unavailable\"t\n\020application/json\022`{\"error\": \"individual request 5xx error\", \"message\": \"individual request 5xx error\", \"code\": 14}Ji\n\003504\022b\n\021Gateway Timed Out\"M\n\020application/json\0229{\"error\": \"Timed out\", \"message\": \"Timed out\", \"code\": 4}'
  _FEATURESERVICETYPE._serialized_start=8337
  _FEATURESERVICETYPE._serialized_end=8384
  _FEATURESERVERDATATYPE._serialized_start=8387
  _FEATURESERVERDATATYPE._serialized_end=8531
  _STATUS._serialized_start=8533
  _STATUS._serialized_end=8592
  _SLOINELIGIBILITYREASON._serialized_start=8595
  _SLOINELIGIBILITYREASON._serialized_end=8749
  _NAMEANDTYPE._serialized_start=294
  _NAMEANDTYPE._serialized_end=613
  _GETFEATURESRESULT._serialized_start=616
  _GETFEATURESRESULT._serialized_end=848
  _FEATURESERVERCOMPLEXDATATYPE._serialized_start=851
  _FEATURESERVERCOMPLEXDATATYPE._serialized_end=1545
  _FEATURESERVERSTRUCTFIELD._serialized_start=1548
  _FEATURESERVERSTRUCTFIELD._serialized_end=2116
  _SLOINFO._serialized_start=2119
  _SLOINFO._serialized_end=2880
  _METADATA._serialized_start=2883
  _METADATA._serialized_end=5273
  _METADATA_ITEM._serialized_start=3162
  _METADATA_ITEM._serialized_end=3988
  _GETFEATURESRESPONSE._serialized_start=5276
  _GETFEATURESRESPONSE._serialized_end=5446
  _GETFEATURESBATCHRESPONSE._serialized_start=5449
  _GETFEATURESBATCHRESPONSE._serialized_end=5635
  _BATCHSLOINFO._serialized_start=5638
  _BATCHSLOINFO._serialized_end=5952
  _BATCHMETADATA._serialized_start=5955
  _BATCHMETADATA._serialized_end=6552
  _BATCHMETADATA_ITEM._serialized_start=6294
  _BATCHMETADATA_ITEM._serialized_end=6552
  _SERVICEMETADATAREQUEST._serialized_start=6554
  _SERVICEMETADATAREQUEST._serialized_end=6659
  _SERVICEMETADATARESPONSE._serialized_start=6662
  _SERVICEMETADATARESPONSE._serialized_end=7424
  _QUERYFEATURESREQUEST._serialized_start=7426
  _QUERYFEATURESREQUEST._serialized_end=7529
  _QUERYFEATURESRESPONSE._serialized_start=7532
  _QUERYFEATURESRESPONSE._serialized_end=7706
  _FEATURESERVICELOCATOR._serialized_start=7709
  _FEATURESERVICELOCATOR._serialized_end=8148
  _GETFEATURESERVICESTATEREQUEST._serialized_start=8151
  _GETFEATURESERVICESTATEREQUEST._serialized_end=8295
  _GETFEATURESERVICESTATERESPONSE._serialized_start=8297
  _GETFEATURESERVICESTATERESPONSE._serialized_end=8335
  _FEATURESERVICE._serialized_start=8752
  _FEATURESERVICE._serialized_end=14625
# @@protoc_insertion_point(module_scope)



import json
import sys

import click

from tecton._internals import metadata_service
from tecton._internals.display import Displayable
from tecton.cli import printer
from tecton.cli.command import TectonGroup
from tecton_proto.auth.authorization_service_pb2 import Assignment
from tecton_proto.auth.authorization_service_pb2 import AssignRolesRequest
from tecton_proto.auth.authorization_service_pb2 import GetAssignedRolesRequest
from tecton_proto.auth.authorization_service_pb2 import GetRolesRequest
from tecton_proto.auth.authorization_service_pb2 import UnassignRolesRequest
from tecton_proto.auth.principal_pb2 import PrincipalType
from tecton_proto.auth.resource_pb2 import ResourceType
from tecton_proto.metadataservice.metadata_service_pb2 import GetUserRequest


RESOURCE_TYPES = {
    "workspace": ResourceType.RESOURCE_TYPE_WORKSPACE,
    "organization": ResourceType.RESOURCE_TYPE_ORGANIZATION,
}


def _get_role_definitions():
    request = GetRolesRequest()
    response = metadata_service.instance().GetRoles(request)
    return response.roles


@click.command("access-control", cls=TectonGroup)
def access_control():
    """Manage Access Controls"""


@access_control.command("assign-role")
@click.option("-w", "--workspace", required=False)
# we can't make the role help dynamic without making top level usage of the CLI make a network request
# since even lazy loading following https://github.com/pallets/click/pull/2348 doesn't work for help text
@click.option(
    "-r", "--role", required=True, type=str, help="Role name (e.g. admin, owner, editor, consumer, viewer, etc)"
)
@click.option("-u", "--user", default=None, help="User Email")
@click.option("-s", "--service-account", default=None, help="Service Account ID")
def assign_role_command(workspace, role, user, service_account):
    """Assign a role to a principal."""
    _update_role(workspace, role, user, service_account)


@access_control.command()
@click.option("-w", "--workspace", required=False)
# we can't make the role help dynamic without making top level usage of the CLI make a network request
# since even lazy loading following https://github.com/pallets/click/pull/2348 doesn't work for help text
@click.option(
    "-r", "--role", required=True, type=str, help="Role name (e.g. admin, owner, editor, consumer, viewer, etc)"
)
@click.option("-u", "--user", default=None, help="User Email")
@click.option("-s", "--service-account", default=None, help="Service Account ID")
def unassign_role(workspace, role, user, service_account):
    """Unassign a role from a principal."""
    _update_role(workspace, role, user, service_account, unassign=True)


def _update_role(workspace, role, user, service_account, unassign=False):
    role = role.lower()
    assignment = Assignment()
    principal_type, principal_id = get_principal_details(user, service_account)

    if role == "admin":
        if workspace:
            raise click.ClickException("'Admin' is a cluster-wide role. Please remove the --workspace argument.")
        resource_type = ResourceType.RESOURCE_TYPE_ORGANIZATION
    elif workspace:
        resource_type = ResourceType.RESOURCE_TYPE_WORKSPACE
        assignment.resource_id = workspace
    else:
        raise click.ClickException("Please mention a workspace name using --workspace")

    assignment.resource_type = resource_type
    assignment.principal_type = principal_type
    assignment.principal_id = principal_id

    role_defs = _get_role_definitions()

    role_def = next((r for r in role_defs if r.id == role), None)
    if role_def is None:
        raise click.ClickException(
            f"Invalid role id. Possible values are: {', '.join(r.id for r in role_defs if _is_role_assignable(r, principal_type, resource_type))}"
        )

    assignment.role = role_def.legacy_id

    try:
        if unassign:
            request = UnassignRolesRequest()
            request.assignments.append(assignment)
            metadata_service.instance().UnassignRoles(request)
        else:
            request = AssignRolesRequest()
            request.assignments.append(assignment)
            metadata_service.instance().AssignRoles(request)
        printer.safe_print("Successfully updated role.")
    except Exception as e:
        printer.safe_print(f"Failed to update role: {e}", file=sys.stderr)
        sys.exit(1)


def _is_role_assignable(role_def, principal_type, resource_type):
    return (
        principal_type in role_def.assignable_to_principal_types
        and resource_type in role_def.assignable_on_resource_types
    )


def get_roles(principal_type, principal_id, resource_type):
    request = GetAssignedRolesRequest()
    request.principal_type = principal_type
    request.principal_id = principal_id
    request.resource_type = resource_type
    response = metadata_service.instance().GetAssignedRoles(request)
    return response


def display_table(headings, ws_roles):
    table = Displayable.from_table(headings=headings, rows=ws_roles, max_width=0)
    # Align columns in the middle horizontally
    table._text_table.set_cols_align(["c" for _ in range(len(headings))])
    printer.safe_print(table)


@access_control.command("get-roles")
@click.option("-u", "--user", default=None, help="User Email")
@click.option("-s", "--service-account", default=None, help="Service Account ID")
@click.option(
    "-r",
    "--resource_type",
    default=None,
    type=click.Choice(RESOURCE_TYPES.keys()),
    help="Optional Resource Type to which the Principal has roles assigned.",
)
@click.option("--json-out", default=False, is_flag=True, help="Format Output as JSON")
def get_assigned_roles(user, service_account, resource_type, json_out):
    """Get the roles assigned to a principal."""
    if resource_type is not None:
        resource_type = RESOURCE_TYPES[resource_type]
    principal_type, principal_id = get_principal_details(user, service_account)

    role_defs = _get_role_definitions()

    try:
        json_output = []
        if resource_type is None or resource_type == ResourceType.RESOURCE_TYPE_WORKSPACE:
            ws_response = get_roles(principal_type, principal_id, ResourceType.RESOURCE_TYPE_WORKSPACE)
            ws_roles = []
            for assignment in ws_response.assignments:
                if len(assignment.roles) > 0:
                    roles = [_maybe_convert_legacy_role_id(role_defs, role) for role in assignment.roles]
                    if json_out:
                        json_output.append(
                            {"resource_type": "WORKSPACE", "workspace_name": assignment.resource_id, "roles": roles}
                        )
                    else:
                        ws_roles.append((assignment.resource_id, ", ".join(roles)))
            if not json_out:
                headings = ["Workspace", "Role"]
                display_table(headings, ws_roles)
                printer.safe_print()

        if resource_type is None or resource_type == ResourceType.RESOURCE_TYPE_ORGANIZATION:
            org_response = get_roles(principal_type, principal_id, ResourceType.RESOURCE_TYPE_ORGANIZATION)
            org_roles = []
            for assignment in org_response.assignments:
                if len(assignment.roles) > 0:
                    roles = [_maybe_convert_legacy_role_id(role_defs, role) for role in assignment.roles]
                    if json_out:
                        json_output.append({"resource_type": "ORGANIZATION", "roles": roles})
                    else:
                        org_roles.append((", ".join(roles),))
            if not json_out:
                headings = ["Organization Roles"]
                display_table(headings, org_roles)

        if json_out:
            printer.safe_print(json.dumps(json_output, indent=4))

    except Exception as e:
        printer.safe_print(f"Failed to Get Roles: {e}", file=sys.stderr)
        sys.exit(1)


def _maybe_convert_legacy_role_id(role_defs, id):
    role_def = next((r for r in role_defs if r.id == id or r.legacy_id == id), None)
    role_id = id if role_def is None else role_def.id
    return role_id


def get_user_id(email):
    try:
        request = GetUserRequest()
        request.email = email
        response = metadata_service.instance().GetUser(request)
        return response.user.okta_id
    except Exception as e:
        printer.safe_print(f"Failed to Get Roles: {e}", file=sys.stderr)
        sys.exit(1)


def get_principal_details(user, service_account):
    if user and service_account:
        raise click.ClickException("Please mention a single Principal Type using one of --user or --service-account")
    if user:
        return PrincipalType.PRINCIPAL_TYPE_USER, get_user_id(user)
    elif service_account:
        return PrincipalType.PRINCIPAL_TYPE_SERVICE_ACCOUNT, service_account
    else:
        raise click.ClickException("Please mention a Principal Type using --user or --service-account")



DEFAULT_AGGREGATION_TILES_WINDOW_START_COLUMN_NAME = "tile_start_time"
DEFAULT_AGGREGATION_TILES_WINDOW_END_COLUMN_NAME = "tile_end_time"

AGGREGATION_LEVEL_DISABLED = "disabled"
AGGREGATION_LEVEL_PARTIAL = "partial"
AGGREGATION_LEVEL_FULL = "full"
SUPPORTED_AGGREGATION_LEVEL_VALUES = [
    AGGREGATION_LEVEL_DISABLED,
    AGGREGATION_LEVEL_PARTIAL,
    AGGREGATION_LEVEL_FULL,
]



import inspect
import threading
from functools import wraps
from typing import Callable
from typing import Optional

import pendulum
from typeguard import typechecked

from tecton._internals import errors
from tecton._internals import metadata_service
from tecton._internals.analytics import AnalyticsLogger
from tecton._internals.metadata_service_impl import trace as metadata_service_trace
from tecton.framework import validation_mode
from tecton_core import conf
from tecton_core.errors import TectonAPIInaccessibleError
from tecton_core.errors import TectonAPIValidationError
from tecton_core.errors import TectonInternalError
from tecton_core.id_helper import IdHelper


analytics = AnalyticsLogger()
thread_local_data = threading.local()
sdk_public_method_decorator_enabled = True


def disable_sdk_public_method_decorator():
    global sdk_public_method_decorator_enabled
    sdk_public_method_decorator_enabled = False


def enable_sdk_public_method_decorator():
    global sdk_public_method_decorator_enabled
    sdk_public_method_decorator_enabled = True


def documented_by(func):
    """Normally you should just use functools.wraps over this, but wraps() doesn't work nicely
    when you try to wrap a classmethod in a normal func.
    """

    def wrapper(target):
        target.__doc__ = func.__doc__
        return target

    return wrapper


def assert_valid(func):
    """Check if the object has been validated, if not throw an error.

    This should only be used for private methods, as opposed to public methods, which should use
    @sdk_public_method(requires_validation=True) and will trigger automatic validation if needed.
    """

    @wraps(func)
    def wrapper(fco_object, *args, **kwargs):
        if not fco_object._is_valid:
            raise errors.TECTON_OBJECT_REQUIRES_VALIDATION(
                func.__name__, type(fco_object).__name__, fco_object.info.name
            )
        return func(fco_object, *args, **kwargs)

    return wrapper


def assert_remote_object(
    original_function: Optional[Callable] = None,
    *,
    error_message: Callable[[str], Exception] = errors.INVALID_USAGE_FOR_LOCAL_TECTON_OBJECT,
):
    """Assert this function is being called on a remote Tecton object, aka an object applied and fetched from the backend, and raise error otherwise.

    :param: error_message: error message to raise if the Tecton object is locally defined. The error_message param must contain a function that takes
    in the target function's name as a param and returns an Exception.
    """

    def inner_decorater(target):
        @wraps(target)
        def wrapper(fco_object, *args, **kwargs):
            if fco_object.info._is_local_object:
                raise error_message(target.__name__)
            return target(fco_object, *args, **kwargs)

        return wrapper

    if original_function:
        return inner_decorater(original_function)

    return inner_decorater


def assert_local_object(
    original_function: Optional[Callable] = None,
    *,
    error_message: Callable[[str], Exception] = errors.INVALID_USAGE_FOR_REMOTE_TECTON_OBJECT,
):
    """Assert this function is being called on a local Tecton object, aka an object created locally (as opposed to being fetched from the backend).

    :param: error_message: error message to raise if the Tecton object is not locally defined. The error_message param must contain a function that takes
    in the target function's name as a param and returns an Exception.
    """

    def inner_decorater(target):
        @wraps(target)
        def wrapper(fco_object, *args, **kwargs):
            if not fco_object.info._is_local_object:
                raise error_message(target.__name__)
            return target(fco_object, *args, **kwargs)

        return wrapper

    if original_function:
        return inner_decorater(original_function)

    return inner_decorater


def sdk_public_method(
    original_function: Optional[Callable] = None,
    *,
    requires_validation: bool = False,
    validation_error_message: Callable[[str, str, str], Exception] = errors.TECTON_OBJECT_REQUIRES_VALIDATION,
):
    """Decorator for public SDK methods that should be have analytics logging.

    :param original_function: The function to be wrapped.
    :param requires_validation: If True, will run Tecton object validation (if needed) before entering running original
        function. Automatically triggered validation will be logged separately. Should only be set to True when used
        to wrap a "Tecton object" (e.g. a DataSource or FeatureView) method.
    :param validation_error_message: The error message function to use if the object has not been validated and
        automatic validation is not enabled.
    """

    def inner_decorator(target_func):
        arg_names = _get_arg_names(target_func)

        @wraps(target_func)
        def _sdk_public_method_wrapper(*args, **kwargs):
            if requires_validation:
                if not _is_method(target_func) or not hasattr(args[0], "_is_valid"):
                    raise TectonInternalError(
                        "sdk_public_method(requires_validation=True) should only be used with Tecton object methods."
                    )
                fco_object = args[0]
                # Object has not been validated yet.
                if not fco_object._is_valid:
                    if conf.get_or_none("TECTON_VALIDATION_MODE") == validation_mode.ValidationMode.AUTOMATIC:
                        fco_object._run_automatic_validation()
                    else:
                        raise validation_error_message(
                            target_func.__name__, type(fco_object).__name__, fco_object.info.name
                        )

            if not sdk_public_method_decorator_enabled:
                return target_func(*args, **kwargs)

            if not hasattr(thread_local_data, "in_tecton_sdk_public_method_wrapper"):
                thread_local_data.in_tecton_sdk_public_method_wrapper = False

            already_in_wrapper = thread_local_data.in_tecton_sdk_public_method_wrapper
            if already_in_wrapper:
                return target_func(*args, **kwargs)

            try:
                thread_local_data.in_tecton_sdk_public_method_wrapper = True
                return _invoke_and_transform_errors(
                    target_func, args, kwargs, arg_names, analytics is not None, not already_in_wrapper, typecheck=True
                )
            finally:
                thread_local_data.in_tecton_sdk_public_method_wrapper = already_in_wrapper

        return _sdk_public_method_wrapper

    if original_function:
        return inner_decorator(original_function)

    return inner_decorator


def _is_method(func):
    params = inspect.signature(func).parameters
    return "self" in params or "cls" in params


def _get_arg_names(func):
    arg_names = []
    for i, param in enumerate(inspect.signature(func).parameters.values()):
        name = param.name
        if i == 0 and _is_method(func):
            continue
        arg_names.append(name)
    return arg_names


def _invoke_and_transform_errors(
    func, args, kwargs, arg_names, log_analytics: bool, is_top_level: bool, typecheck: bool = False
):
    if typecheck:
        func = typechecked(func)

    if _is_method(func):
        method_object = args[0]
        method_args = args[1:]
    else:
        method_object = None
        method_args = args

    trace_id = IdHelper.generate_string_id()

    if log_analytics:
        start_time = pendulum.now("UTC")
        analytics.log_method_entry(trace_id, method_object, func, method_args, kwargs, arg_names)

    metadata_service_trace.set_trace_id(trace_id)
    return_value = None

    try:
        return_value = func(*args, **kwargs)
        caught_exception = None
        return return_value
    except TectonAPIValidationError as e:
        caught_exception = e
        raise errors.VALIDATION_ERROR_FROM_MDS(str(e), trace_id)
    except TectonAPIInaccessibleError as e:
        caught_exception = e
        raise errors.MDS_INACCESSIBLE(metadata_service._get_host_port())
    except Exception as e:
        caught_exception = e
        # Do not chain the exception, just let it pass through. This leads to more legible stack traces and errors.
        raise
    finally:
        metadata_service_trace.set_trace_id(None)
        if log_analytics:
            end_time = pendulum.now("UTC")
            execution_time = end_time - start_time
            analytics.log_method_return(trace_id, method_object, func, return_value, execution_time, caught_exception)



import pyspark.sql.types as spark_types

from tecton_core import schema_derivation_utils as core_schema_derivation_utils
from tecton_core.data_types import ArrayType
from tecton_core.data_types import BoolType
from tecton_core.data_types import DataType
from tecton_core.data_types import Float32Type
from tecton_core.data_types import Float64Type
from tecton_core.data_types import Int32Type
from tecton_core.data_types import Int64Type
from tecton_core.data_types import StringType
from tecton_core.data_types import StructField
from tecton_core.data_types import StructType
from tecton_core.data_types import TimestampType
from tecton_core.schema import Schema
from tecton_proto.common.schema_pb2 import Schema as SchemaProto

# Keep in sync with DataTypeUtils.kt and tecton_core/schema_derivation_utils. . Use "simple strings" as the keys so that fields like "nullable" are ignored.
PRIMITIVE_SPARK_TYPE_SIMPLE_STRING_TO_TECTON_TYPE = {
    spark_types.StringType().simpleString(): StringType(),
    spark_types.LongType().simpleString(): Int64Type(),
    spark_types.FloatType().simpleString(): Float32Type(),
    spark_types.DoubleType().simpleString(): Float64Type(),
    spark_types.BooleanType().simpleString(): BoolType(),
    spark_types.IntegerType().simpleString(): Int32Type(),
    spark_types.TimestampType().simpleString(): TimestampType(),
}

# Map from simple (i.e non-complex) Tecton data types to Spark Types.
PRIMITIVE_TECTON_DATA_TYPE_TO_SPARK_DATA_TYPE = {
    Int32Type(): spark_types.IntegerType(),
    Int64Type(): spark_types.LongType(),
    Float32Type(): spark_types.FloatType(),
    Float64Type(): spark_types.DoubleType(),
    StringType(): spark_types.StringType(),
    BoolType(): spark_types.BooleanType(),
    TimestampType(): spark_types.TimestampType(),
}


def schema_from_spark(spark_schema: spark_types.StructType) -> Schema:
    proto = SchemaProto()
    for field in spark_schema:
        column = proto.columns.add()
        spark_datatype = field.dataType
        try:
            tecton_type = tecton_data_type_from_spark_data_type(spark_datatype)
        except ValueError:
            # Re-raise this error here because
            # a) we have the feature column name and
            # b) in the case of complex data type with a nested type error, can show the full data type.
            raise ValueError(
                f"Field {field.name} is of type {spark_datatype.simpleString()}, which is not a supported type for features. "
                + f"Please change {field.name} to be one of our supported types: https://docs.tecton.ai/latest/faq/creating_and_managing_features.html"
            )
        column.CopyFrom(core_schema_derivation_utils.column_from_tecton_data_type(tecton_type))
        column.name = field.name

    return Schema(proto)


def tecton_data_type_from_spark_data_type(spark_datatype: spark_types.DataType) -> DataType:
    if isinstance(spark_datatype, spark_types.StructType):
        struct_fields = []
        for field in spark_datatype:
            tecton_type = tecton_data_type_from_spark_data_type(field.dataType)
            struct_fields.append(StructField(field.name, tecton_type))
        return StructType(struct_fields)
    elif isinstance(spark_datatype, spark_types.ArrayType):
        return ArrayType(tecton_data_type_from_spark_data_type(spark_datatype.elementType))
    elif spark_datatype.simpleString() in PRIMITIVE_SPARK_TYPE_SIMPLE_STRING_TO_TECTON_TYPE:
        return PRIMITIVE_SPARK_TYPE_SIMPLE_STRING_TO_TECTON_TYPE[spark_datatype.simpleString()]
    else:
        raise ValueError(f"{spark_datatype.simpleString()}, is not a supported type for features.")


def spark_data_type_from_tecton_data_type(tecton_data_type: DataType) -> spark_types.DataType:
    if tecton_data_type in PRIMITIVE_TECTON_DATA_TYPE_TO_SPARK_DATA_TYPE:
        return PRIMITIVE_TECTON_DATA_TYPE_TO_SPARK_DATA_TYPE[tecton_data_type]
    elif isinstance(tecton_data_type, ArrayType):
        element_type = spark_data_type_from_tecton_data_type(tecton_data_type.element_type)
        return spark_types.ArrayType(element_type)
    elif isinstance(tecton_data_type, StructType):
        spark_struct = spark_types.StructType()
        for field in tecton_data_type.fields:
            spark_struct.add(field.name, spark_data_type_from_tecton_data_type(field.data_type))
        return spark_struct
    else:
        assert False, f"Unsupported type: {tecton_data_type}"


def schema_to_spark(schema: Schema) -> spark_types.StructType:
    ret = spark_types.StructType()
    for col_name, col_spark_data_type in column_name_spark_data_types(schema):
        ret.add(col_name, col_spark_data_type)
    return ret


def column_name_spark_data_types(schema: Schema):
    return [(c[0], spark_data_type_from_tecton_data_type(c[1])) for c in schema.column_name_and_data_types()]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/saved_feature_data_frame.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from tecton_proto.common import id_pb2 as tecton__proto_dot_common_dot_id__pb2
from tecton_proto.common import spark_schema_pb2 as tecton__proto_dot_common_dot_spark__schema__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n0tecton_proto/data/saved_feature_data_frame.proto\x12\x11tecton_proto.data\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1ctecton_proto/common/id.proto\x1a&tecton_proto/common/spark_schema.proto\"\xd1\x01\n\x19SavedFeatureDataFrameInfo\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x39\n\ncreated_at\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tcreatedAt\x12\x1f\n\x0bis_archived\x18\x03 \x01(\x08R\nisArchived\x12\x1c\n\tworkspace\x18\x04 \x01(\tR\tworkspace\x12&\n\x0fis_data_deleted\x18\x05 \x01(\x08R\risDataDeleted\"\x9f\x06\n\x15SavedFeatureDataFrame\x12T\n\x1asaved_feature_dataframe_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x17savedFeatureDataframeId\x12@\n\x04info\x18\x02 \x01(\x0b\x32,.tecton_proto.data.SavedFeatureDataFrameInfoR\x04info\x12-\n\x12\x64\x61taframe_location\x18\x03 \x01(\tR\x11\x64\x61taframeLocation\x12G\n\x12\x66\x65\x61ture_package_id\x18\x04 \x01(\x0b\x32\x17.tecton_proto.common.IdH\x00R\x10\x66\x65\x61turePackageId\x12G\n\x12\x66\x65\x61ture_service_id\x18\x05 \x01(\x0b\x32\x17.tecton_proto.common.IdH\x00R\x10\x66\x65\x61tureServiceId\x12\x32\n\x14\x66\x65\x61ture_package_name\x18\x06 \x01(\tH\x01R\x12\x66\x65\x61turePackageName\x12\x32\n\x14\x66\x65\x61ture_service_name\x18\x07 \x01(\tH\x01R\x12\x66\x65\x61tureServiceName\x12>\n\x1cstate_update_entry_commit_id\x18\x08 \x01(\tR\x18stateUpdateEntryCommitId\x12\x31\n\x15join_key_column_names\x18\t \x03(\tR\x12joinKeyColumnNames\x12\x32\n\x15timestamp_column_name\x18\n \x01(\tR\x13timestampColumnName\x12\x38\n\x06schema\x18\x0b \x01(\x0b\x32 .tecton_proto.common.SparkSchemaR\x06schema\x12@\n\x04type\x18\x0c \x01(\x0e\x32,.tecton_proto.data.SavedFeatureDataFrameTypeR\x04typeB\x0f\n\rsource_fco_idB\x11\n\x0fsource_fco_name*?\n\x19SavedFeatureDataFrameType\x12\x0b\n\x07NOT_SET\x10\x00\x12\t\n\x05SAVED\x10\x01\x12\n\n\x06LOGGED\x10\x02\x42\x13\n\x0f\x63om.tecton.dataP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.data.saved_feature_data_frame_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.dataP\001'
  _SAVEDFEATUREDATAFRAMETYPE._serialized_start=1188
  _SAVEDFEATUREDATAFRAMETYPE._serialized_end=1251
  _SAVEDFEATUREDATAFRAMEINFO._serialized_start=175
  _SAVEDFEATUREDATAFRAMEINFO._serialized_end=384
  _SAVEDFEATUREDATAFRAME._serialized_start=387
  _SAVEDFEATUREDATAFRAME._serialized_end=1186
# @@protoc_insertion_point(module_scope)



from typing import Optional

import attr
import pendulum
from pyspark.sql import DataFrame
from pyspark.sql import functions
from pyspark.sql import SparkSession

from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper as FeatureDefinition
from tecton_core.logger import get_logger
from tecton_core.query.builder import build_materialization_querytree
from tecton_core.query.node_interface import NodeRef
from tecton_core.query.rewrite import rewrite_tree
from tecton_core.query_consts import ANCHOR_TIME
from tecton_core.time_utils import convert_timedelta_for_version
from tecton_spark.data_observability import MetricsCollector
from tecton_spark.query import translate

MATERIALIZED_RAW_DATA_END_TIME = "_materialized_raw_data_end_time"
logger = get_logger("MaterializationPlan")


@attr.s(auto_attribs=True)
class MaterializationPlan:
    """Computes dataframes required for materialization to offline and online stores.

    This class contains `offline_store_data_frame` and `online_store_data_frame`, but both might not always be necessary.

    Both `offline_store_data_frame` and `online_store_data_frame` should be used during batch materialization of BFVs and BWAFVs.
    Only `offline_store_data_frame` should be used during batch materialization of SFVs and SWAFVs.
    Only `online_store_data_frame` should be used during stream materialization of SFVs and SWAFVs.
    """

    _fd: FeatureDefinition
    base_data_frame: DataFrame

    @classmethod
    def from_querytree(cls, fd: FeatureDefinition, query_tree: NodeRef, spark: SparkSession) -> "MaterializationPlan":
        logger.info("Query Plan:")
        logger.info(query_tree.pretty_str())
        spark_df = translate.spark_convert(query_tree).to_dataframe(spark)
        return cls(fd=fd, base_data_frame=spark_df)

    @classmethod
    def from_parquet(cls, fd: FeatureDefinition, path: str, spark: SparkSession):
        return cls(fd=fd, base_data_frame=spark.read.parquet(path))

    @property
    def offline_store_data_frame(self) -> DataFrame:
        return self.base_data_frame

    @property
    def online_store_data_frame(self) -> DataFrame:
        online_df = self.base_data_frame

        # batch online and offline df are slightly different
        if self._fd.is_temporal and not online_df.isStreaming:
            version = self._fd.get_feature_store_format_version
            batch_mat_schedule = convert_timedelta_for_version(self._fd.batch_materialization_schedule, version)
            online_df = self.base_data_frame.withColumn(
                MATERIALIZED_RAW_DATA_END_TIME, functions.col(ANCHOR_TIME) + batch_mat_schedule
            ).drop(ANCHOR_TIME)
        return online_df


def get_batch_materialization_plan(
    *,
    spark: SparkSession,
    feature_definition: FeatureDefinition,
    feature_data_time_limits: Optional[pendulum.Period],
    metrics_collector: Optional[MetricsCollector] = None,
) -> MaterializationPlan:
    """
    NOTE: We rely on Spark's lazy evaluation model to infer partially materialized tile Schema during FeatureView
    creation time without actually performing any materialization.
    Please make sure to not perform any Spark operations under this function's code path that will actually execute
    the Spark query (e.g: df.count(), df.show(), etc.).
    """
    query_tree = build_materialization_querytree(
        feature_definition,
        for_stream=False,
        feature_data_time_limits=feature_data_time_limits,
        enable_feature_metrics=(metrics_collector is not None),
    )
    rewrite_tree(query_tree)
    return MaterializationPlan.from_querytree(fd=feature_definition, query_tree=query_tree, spark=spark)


def get_stream_materialization_plan(
    *,
    spark: SparkSession,
    feature_definition: FeatureDefinition,
    metrics_collector: Optional[MetricsCollector] = None,
) -> MaterializationPlan:
    query_tree = build_materialization_querytree(
        feature_definition, for_stream=True, enable_feature_metrics=(metrics_collector is not None)
    )
    rewrite_tree(query_tree)
    return MaterializationPlan.from_querytree(fd=feature_definition, query_tree=query_tree, spark=spark)



from tecton_core.id_helper import IdHelper
from tecton_proto.common import id_pb2

_ALL_FCOS = {}


# An internal superclass of declarative Tecton Objects
class Fco:
    @classmethod
    def _register(cls, fco):
        id = fco._id
        assert id.most_significant_bits > 0 and id.least_significant_bits > 0, f"Object {fco} has no `id` set."
        _ALL_FCOS[IdHelper.to_string(id)] = fco

    @property
    def _id(self) -> id_pb2.Id:
        """
        The id of this Tecton Object.
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/hive_metastore.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n&tecton_proto/data/hive_metastore.proto\x12\x11tecton_proto.data\"G\n\x13HiveTableDataSource\x12\x14\n\x05table\x18\x01 \x01(\tR\x05table\x12\x1a\n\x08\x64\x61tabase\x18\x02 \x01(\tR\x08\x64\x61tabase\"&\n\x0eListHiveResult\x12\x14\n\x05names\x18\x01 \x03(\tR\x05namesB\x13\n\x0f\x63om.tecton.dataP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.data.hive_metastore_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.dataP\001'
  _HIVETABLEDATASOURCE._serialized_start=61
  _HIVETABLEDATASOURCE._serialized_end=132
  _LISTHIVERESULT._serialized_start=134
  _LISTHIVERESULT._serialized_end=172
# @@protoc_insertion_point(module_scope)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
User-defined function related classes and functions
"""
import functools
import sys

from pyspark import SparkContext
from pyspark.rdd import _prepare_for_python_RDD, PythonEvalType
from pyspark.sql.column import Column, _to_java_column, _to_seq
from pyspark.sql.types import StringType, DataType, StructType, _parse_datatype_string
from pyspark.sql.pandas.types import to_arrow_type

__all__ = ["UDFRegistration"]


def _wrap_function(sc, func, returnType):
    command = (func, returnType)
    pickled_command, broadcast_vars, env, includes = _prepare_for_python_RDD(sc, command)
    return sc._jvm.PythonFunction(bytearray(pickled_command), env, includes, sc.pythonExec,
                                  sc.pythonVer, broadcast_vars, sc._javaAccumulator)


def _create_udf(f, returnType, evalType):
    # Set the name of the UserDefinedFunction object to be the name of function f
    udf_obj = UserDefinedFunction(
        f, returnType=returnType, name=None, evalType=evalType, deterministic=True)
    return udf_obj._wrapped()


class UserDefinedFunction(object):
    """
    User defined function in Python

    .. versionadded:: 1.3

    Notes
    -----
    The constructor of this class is not supposed to be directly called.
    Use :meth:`pyspark.sql.functions.udf` or :meth:`pyspark.sql.functions.pandas_udf`
    to create this instance.
    """
    def __init__(self, func,
                 returnType=StringType(),
                 name=None,
                 evalType=PythonEvalType.SQL_BATCHED_UDF,
                 deterministic=True):
        if not callable(func):
            raise TypeError(
                "Invalid function: not a function or callable (__call__ is not defined): "
                "{0}".format(type(func)))

        if not isinstance(returnType, (DataType, str)):
            raise TypeError(
                "Invalid return type: returnType should be DataType or str "
                "but is {}".format(returnType))

        if not isinstance(evalType, int):
            raise TypeError(
                "Invalid evaluation type: evalType should be an int but is {}".format(evalType))

        self.func = func
        self._returnType = returnType
        # Stores UserDefinedPythonFunctions jobj, once initialized
        self._returnType_placeholder = None
        self._judf_placeholder = None
        self._name = name or (
            func.__name__ if hasattr(func, '__name__')
            else func.__class__.__name__)
        self.evalType = evalType
        self.deterministic = deterministic

    @property
    def returnType(self):
        # This makes sure this is called after SparkContext is initialized.
        # ``_parse_datatype_string`` accesses to JVM for parsing a DDL formatted string.
        if self._returnType_placeholder is None:
            if isinstance(self._returnType, DataType):
                self._returnType_placeholder = self._returnType
            else:
                self._returnType_placeholder = _parse_datatype_string(self._returnType)

        if self.evalType == PythonEvalType.SQL_SCALAR_PANDAS_UDF or \
                self.evalType == PythonEvalType.SQL_SCALAR_PANDAS_ITER_UDF:
            try:
                to_arrow_type(self._returnType_placeholder)
            except TypeError:
                raise NotImplementedError(
                    "Invalid return type with scalar Pandas UDFs: %s is "
                    "not supported" % str(self._returnType_placeholder))
        elif self.evalType == PythonEvalType.SQL_GROUPED_MAP_PANDAS_UDF:
            if isinstance(self._returnType_placeholder, StructType):
                try:
                    to_arrow_type(self._returnType_placeholder)
                except TypeError:
                    raise NotImplementedError(
                        "Invalid return type with grouped map Pandas UDFs or "
                        "at groupby.applyInPandas: %s is not supported" % str(
                            self._returnType_placeholder))
            else:
                raise TypeError("Invalid return type for grouped map Pandas "
                                "UDFs or at groupby.applyInPandas: return type must be a "
                                "StructType.")
        elif self.evalType == PythonEvalType.SQL_MAP_PANDAS_ITER_UDF:
            if isinstance(self._returnType_placeholder, StructType):
                try:
                    to_arrow_type(self._returnType_placeholder)
                except TypeError:
                    raise NotImplementedError(
                        "Invalid return type in mapInPandas: "
                        "%s is not supported" % str(self._returnType_placeholder))
            else:
                raise TypeError("Invalid return type in mapInPandas: "
                                "return type must be a StructType.")
        elif self.evalType == PythonEvalType.SQL_COGROUPED_MAP_PANDAS_UDF:
            if isinstance(self._returnType_placeholder, StructType):
                try:
                    to_arrow_type(self._returnType_placeholder)
                except TypeError:
                    raise NotImplementedError(
                        "Invalid return type in cogroup.applyInPandas: "
                        "%s is not supported" % str(self._returnType_placeholder))
            else:
                raise TypeError("Invalid return type in cogroup.applyInPandas: "
                                "return type must be a StructType.")
        elif self.evalType == PythonEvalType.SQL_GROUPED_AGG_PANDAS_UDF:
            try:
                # StructType is not yet allowed as a return type, explicitly check here to fail fast
                if isinstance(self._returnType_placeholder, StructType):
                    raise TypeError
                to_arrow_type(self._returnType_placeholder)
            except TypeError:
                raise NotImplementedError(
                    "Invalid  return type with grouped aggregate Pandas UDFs: "
                    "%s is not supported" % str(self._returnType_placeholder))

        return self._returnType_placeholder

    @property
    def _judf(self):
        # It is possible that concurrent access, to newly created UDF,
        # will initialize multiple UserDefinedPythonFunctions.
        # This is unlikely, doesn't affect correctness,
        # and should have a minimal performance impact.
        if self._judf_placeholder is None:
            self._judf_placeholder = self._create_judf()
        return self._judf_placeholder

    def _create_judf(self):
        from pyspark.sql import SparkSession

        spark = SparkSession.builder.getOrCreate()
        sc = spark.sparkContext

        wrapped_func = _wrap_function(sc, self.func, self.returnType)
        jdt = spark._jsparkSession.parseDataType(self.returnType.json())
        judf = sc._jvm.org.apache.spark.sql.execution.python.UserDefinedPythonFunction(
            self._name, wrapped_func, jdt, self.evalType, self.deterministic)
        return judf

    def __call__(self, *cols):
        judf = self._judf
        sc = SparkContext._active_spark_context
        return Column(judf.apply(_to_seq(sc, cols, _to_java_column)))

    # This function is for improving the online help system in the interactive interpreter.
    # For example, the built-in help / pydoc.help. It wraps the UDF with the docstring and
    # argument annotation. (See: SPARK-19161)
    def _wrapped(self):
        """
        Wrap this udf with a function and attach docstring from func
        """

        # It is possible for a callable instance without __name__ attribute or/and
        # __module__ attribute to be wrapped here. For example, functools.partial. In this case,
        # we should avoid wrapping the attributes from the wrapped function to the wrapper
        # function. So, we take out these attribute names from the default names to set and
        # then manually assign it after being wrapped.
        assignments = tuple(
            a for a in functools.WRAPPER_ASSIGNMENTS if a != '__name__' and a != '__module__')

        @functools.wraps(self.func, assigned=assignments)
        def wrapper(*args):
            return self(*args)

        wrapper.__name__ = self._name
        wrapper.__module__ = (self.func.__module__ if hasattr(self.func, '__module__')
                              else self.func.__class__.__module__)

        wrapper.func = self.func
        wrapper.returnType = self.returnType
        wrapper.evalType = self.evalType
        wrapper.deterministic = self.deterministic
        wrapper.asNondeterministic = functools.wraps(
            self.asNondeterministic)(lambda: self.asNondeterministic()._wrapped())
        return wrapper

    def asNondeterministic(self):
        """
        Updates UserDefinedFunction to nondeterministic.

        .. versionadded:: 2.3
        """
        # Here, we explicitly clean the cache to create a JVM UDF instance
        # with 'deterministic' updated. See SPARK-23233.
        self._judf_placeholder = None
        self.deterministic = False
        return self


class UDFRegistration(object):
    """
    Wrapper for user-defined function registration. This instance can be accessed by
    :attr:`spark.udf` or :attr:`sqlContext.udf`.

    .. versionadded:: 1.3.1
    """

    def __init__(self, sparkSession):
        self.sparkSession = sparkSession

    def register(self, name, f, returnType=None):
        """Register a Python function (including lambda function) or a user-defined function
        as a SQL function.

        .. versionadded:: 1.3.1

        Parameters
        ----------
        name : str,
            name of the user-defined function in SQL statements.
        f : function, :meth:`pyspark.sql.functions.udf` or :meth:`pyspark.sql.functions.pandas_udf`
            a Python function, or a user-defined function. The user-defined function can
            be either row-at-a-time or vectorized. See :meth:`pyspark.sql.functions.udf` and
            :meth:`pyspark.sql.functions.pandas_udf`.
        returnType : :class:`pyspark.sql.types.DataType` or str, optional
            the return type of the registered user-defined function. The value can
            be either a :class:`pyspark.sql.types.DataType` object or a DDL-formatted type string.
            `returnType` can be optionally specified when `f` is a Python function but not
            when `f` is a user-defined function. Please see the examples below.

        Returns
        -------
        function
            a user-defined function

        Notes
        -----
        To register a nondeterministic Python function, users need to first build
        a nondeterministic user-defined function for the Python function and then register it
        as a SQL function.

        Examples
        --------
        1. When `f` is a Python function:

            `returnType` defaults to string type and can be optionally specified. The produced
            object must match the specified type. In this case, this API works as if
            `register(name, f, returnType=StringType())`.

            >>> strlen = spark.udf.register("stringLengthString", lambda x: len(x))
            >>> spark.sql("SELECT stringLengthString('test')").collect()
            [Row(stringLengthString(test)='4')]

            >>> spark.sql("SELECT 'foo' AS text").select(strlen("text")).collect()
            [Row(stringLengthString(text)='3')]

            >>> from pyspark.sql.types import IntegerType
            >>> _ = spark.udf.register("stringLengthInt", lambda x: len(x), IntegerType())
            >>> spark.sql("SELECT stringLengthInt('test')").collect()
            [Row(stringLengthInt(test)=4)]

            >>> from pyspark.sql.types import IntegerType
            >>> _ = spark.udf.register("stringLengthInt", lambda x: len(x), IntegerType())
            >>> spark.sql("SELECT stringLengthInt('test')").collect()
            [Row(stringLengthInt(test)=4)]

        2. When `f` is a user-defined function (from Spark 2.3.0):

            Spark uses the return type of the given user-defined function as the return type of
            the registered user-defined function. `returnType` should not be specified.
            In this case, this API works as if `register(name, f)`.

            >>> from pyspark.sql.types import IntegerType
            >>> from pyspark.sql.functions import udf
            >>> slen = udf(lambda s: len(s), IntegerType())
            >>> _ = spark.udf.register("slen", slen)
            >>> spark.sql("SELECT slen('test')").collect()
            [Row(slen(test)=4)]

            >>> import random
            >>> from pyspark.sql.functions import udf
            >>> from pyspark.sql.types import IntegerType
            >>> random_udf = udf(lambda: random.randint(0, 100), IntegerType()).asNondeterministic()
            >>> new_random_udf = spark.udf.register("random_udf", random_udf)
            >>> spark.sql("SELECT random_udf()").collect()  # doctest: +SKIP
            [Row(random_udf()=82)]

            >>> import pandas as pd  # doctest: +SKIP
            >>> from pyspark.sql.functions import pandas_udf
            >>> @pandas_udf("integer")  # doctest: +SKIP
            ... def add_one(s: pd.Series) -> pd.Series:
            ...     return s + 1
            ...
            >>> _ = spark.udf.register("add_one", add_one)  # doctest: +SKIP
            >>> spark.sql("SELECT add_one(id) FROM range(3)").collect()  # doctest: +SKIP
            [Row(add_one(id)=1), Row(add_one(id)=2), Row(add_one(id)=3)]

            >>> @pandas_udf("integer")  # doctest: +SKIP
            ... def sum_udf(v: pd.Series) -> int:
            ...     return v.sum()
            ...
            >>> _ = spark.udf.register("sum_udf", sum_udf)  # doctest: +SKIP
            >>> q = "SELECT sum_udf(v1) FROM VALUES (3, 0), (2, 0), (1, 1) tbl(v1, v2) GROUP BY v2"
            >>> spark.sql(q).collect()  # doctest: +SKIP
            [Row(sum_udf(v1)=1), Row(sum_udf(v1)=5)]

        """

        # This is to check whether the input function is from a user-defined function or
        # Python function.
        if hasattr(f, 'asNondeterministic'):
            if returnType is not None:
                raise TypeError(
                    "Invalid return type: data type can not be specified when f is"
                    "a user-defined function, but got %s." % returnType)
            if f.evalType not in [PythonEvalType.SQL_BATCHED_UDF,
                                  PythonEvalType.SQL_SCALAR_PANDAS_UDF,
                                  PythonEvalType.SQL_SCALAR_PANDAS_ITER_UDF,
                                  PythonEvalType.SQL_GROUPED_AGG_PANDAS_UDF,
                                  PythonEvalType.SQL_MAP_PANDAS_ITER_UDF]:
                raise ValueError(
                    "Invalid f: f must be SQL_BATCHED_UDF, SQL_SCALAR_PANDAS_UDF, "
                    "SQL_SCALAR_PANDAS_ITER_UDF, SQL_GROUPED_AGG_PANDAS_UDF or "
                    "SQL_MAP_PANDAS_ITER_UDF.")
            register_udf = UserDefinedFunction(f.func, returnType=f.returnType, name=name,
                                               evalType=f.evalType,
                                               deterministic=f.deterministic)
            return_udf = f
        else:
            if returnType is None:
                returnType = StringType()
            register_udf = UserDefinedFunction(f, returnType=returnType, name=name,
                                               evalType=PythonEvalType.SQL_BATCHED_UDF)
            return_udf = register_udf._wrapped()
        self.sparkSession._jsparkSession.udf().registerPython(name, register_udf._judf)
        return return_udf

    def registerJavaFunction(self, name, javaClassName, returnType=None):
        """Register a Java user-defined function as a SQL function.

        In addition to a name and the function itself, the return type can be optionally specified.
        When the return type is not specified we would infer it via reflection.

        .. versionadded:: 2.3.0

        Parameters
        ----------
        name : str
            name of the user-defined function
        javaClassName : str
            fully qualified name of java class
        returnType : :class:`pyspark.sql.types.DataType` or str, optional
            the return type of the registered Java function. The value can be either
            a :class:`pyspark.sql.types.DataType` object or a DDL-formatted type string.

        Examples
        --------
        >>> from pyspark.sql.types import IntegerType
        >>> spark.udf.registerJavaFunction(
        ...     "javaStringLength", "test.org.apache.spark.sql.JavaStringLength", IntegerType())
        ... # doctest: +SKIP
        >>> spark.sql("SELECT javaStringLength('test')").collect()  # doctest: +SKIP
        [Row(javaStringLength(test)=4)]

        >>> spark.udf.registerJavaFunction(
        ...     "javaStringLength2", "test.org.apache.spark.sql.JavaStringLength")
        ... # doctest: +SKIP
        >>> spark.sql("SELECT javaStringLength2('test')").collect()  # doctest: +SKIP
        [Row(javaStringLength2(test)=4)]

        >>> spark.udf.registerJavaFunction(
        ...     "javaStringLength3", "test.org.apache.spark.sql.JavaStringLength", "integer")
        ... # doctest: +SKIP
        >>> spark.sql("SELECT javaStringLength3('test')").collect()  # doctest: +SKIP
        [Row(javaStringLength3(test)=4)]
        """

        jdt = None
        if returnType is not None:
            if not isinstance(returnType, DataType):
                returnType = _parse_datatype_string(returnType)
            jdt = self.sparkSession._jsparkSession.parseDataType(returnType.json())
        self.sparkSession._jsparkSession.udf().registerJava(name, javaClassName, jdt)

    def registerJavaUDAF(self, name, javaClassName):
        """Register a Java user-defined aggregate function as a SQL function.

        .. versionadded:: 2.3.0

        name : str
            name of the user-defined aggregate function
        javaClassName : str
            fully qualified name of java class

        Examples
        --------
        >>> spark.udf.registerJavaUDAF("javaUDAF", "test.org.apache.spark.sql.MyDoubleAvg")
        ... # doctest: +SKIP
        >>> df = spark.createDataFrame([(1, "a"),(2, "b"), (3, "a")],["id", "name"])
        >>> df.createOrReplaceTempView("df")
        >>> q = "SELECT name, javaUDAF(id) as avg from df group by name order by name desc"
        >>> spark.sql(q).collect()  # doctest: +SKIP
        [Row(name='b', avg=102.0), Row(name='a', avg=102.0)]
        """

        self.sparkSession._jsparkSession.udf().registerJavaUDAF(name, javaClassName)


def _test():
    import doctest
    from pyspark.sql import SparkSession
    import pyspark.sql.udf
    globs = pyspark.sql.udf.__dict__.copy()
    spark = SparkSession.builder\
        .master("local[4]")\
        .appName("sql.udf tests")\
        .getOrCreate()
    globs['spark'] = spark
    (failure_count, test_count) = doctest.testmod(
        pyspark.sql.udf, globs=globs,
        optionflags=doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE)
    spark.stop()
    if failure_count:
        sys.exit(-1)


if __name__ == "__main__":
    _test()



from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import attrs
from typeguard import typechecked

from tecton._internals import display
from tecton._internals import fco as internal_fco
from tecton._internals import metadata_service
from tecton._internals import sdk_decorators
from tecton._internals import validations_api
from tecton.framework import base_tecton_object
from tecton_core import feature_definition_wrapper
from tecton_core import id_helper
from tecton_core import specs
from tecton_proto.args import basic_info_pb2
from tecton_proto.args import entity_pb2 as entity__args_pb2
from tecton_proto.args import fco_args_pb2
from tecton_proto.common import fco_locator_pb2
from tecton_proto.metadataservice import metadata_service_pb2
from tecton_proto.validation import validator_pb2


@attrs.define(eq=False)
class Entity(base_tecton_object.BaseTectonObject, internal_fco.Fco):
    """A Tecton Entity, used to organize and join features.

    An Entity is a class that represents an Entity that is being modeled in Tecton. Entities are used to index and
    organize features - a :class:`FeatureView` contains at least one Entity.

    Entities contain metadata about *join keys*, which represent the columns that are used to join features together.

    Example of an Entity declaration:

    .. code-block:: python

        from tecton import Entity

        customer = Entity(
            name='customer',
            join_keys=['customer_id'],
            description='A customer subscribing to a Sports TV subscription service',
            owner='matt@tecton.ai',
            tags={'release': 'development'}
    """

    # An entity spec, i.e. a dataclass representation of the Tecton object that is used in most functional use cases,
    # e.g. constructing queries. Set only after the object has been validated. Remote objects, i.e. applied objects
    # fetched from the backend, are assumed valid.
    _spec: Optional[specs.EntitySpec] = attrs.field(repr=False)

    # A Tecton "args" proto. Only set if this object was defined locally, i.e. this object was not applied and fetched
    # from the Tecton backend.
    _args: Optional[entity__args_pb2.EntityArgs] = attrs.field(repr=False, on_setattr=attrs.setters.frozen)

    @typechecked
    def __init__(
        self,
        *,
        name: str,
        description: Optional[str] = None,
        tags: Optional[Dict[str, str]] = None,
        owner: Optional[str] = None,
        prevent_destroy: bool = False,
        join_keys: Optional[Union[str, List[str]]] = None,
    ):
        """Declare a new Entity.

        :param name: Unique name for the new entity.
        :param description: Short description of the new entity.
        :param tags: Tags associated with this Tecton Object (key-value pairs of arbitrary metadata).
        :param owner: Owner name (typically the email of the primary maintainer).
        :param prevent_destroy: If True, this Tecton object will be blocked from being deleted or re-created (i.e. a
            destructive update) during tecton plan/apply. To remove or update this object, `prevent_destroy` must be
            first set to False via a separate tecton apply. `prevent_destroy` can be used to prevent accidental changes
            such as inadvertantly deleting a Feature Service used in production or recreating a Feature View that
            triggers expensive rematerialization jobs. `prevent_destroy` also blocks changes to dependent Tecton objects
            that would trigger a recreate of the tagged object, e.g. if `prevent_destroy` is set on a Feature Service,
            that will also prevent deletions or re-creates of Feature Views used in that service. `prevent_destroy` is
            only enforced in live (i.e. non-dev) workspaces.
        :param join_keys: Names of columns that uniquely identify the entity in FeatureView's SQL statement
            for which features should be aggregated. Defaults to using ``name`` as the entity's join key.

        :raises TectonValidationError: if the input non-parameters are invalid.
        """
        from tecton.cli import common as cli_common

        if not join_keys:
            resolved_join_keys = [name]
        elif isinstance(join_keys, str):
            resolved_join_keys = [join_keys]
        else:
            resolved_join_keys = join_keys

        args = entity__args_pb2.EntityArgs(
            entity_id=id_helper.IdHelper.generate_id(),
            info=basic_info_pb2.BasicInfo(name=name, description=description, tags=tags, owner=owner),
            join_keys=resolved_join_keys,
            version=feature_definition_wrapper.FrameworkVersion.FWV5.value,
            prevent_destroy=prevent_destroy,
        )
        info = base_tecton_object.TectonObjectInfo.from_args_proto(args.info, args.entity_id)
        source_info = cli_common.construct_fco_source_info(args.entity_id)
        self.__attrs_init__(info=info, spec=None, args=args, source_info=source_info)
        internal_fco.Fco._register(self)

    @classmethod
    @typechecked
    def _from_spec(cls, spec: specs.EntitySpec) -> "Entity":
        """Create an Entity from directly from a spec. Specs are assumed valid and will not be re-validated."""
        info = base_tecton_object.TectonObjectInfo.from_spec(spec)
        obj = cls.__new__(cls)  # Instantiate the object. Does not call init.
        obj.__attrs_init__(info=info, spec=spec, args=None, source_info=None)
        return obj

    @sdk_decorators.assert_local_object
    def _build_args(self) -> fco_args_pb2.FcoArgs:
        return fco_args_pb2.FcoArgs(entity=self._args)

    def _build_fco_validation_args(self) -> validator_pb2.FcoValidationArgs:
        if self.info._is_local_object:
            return validator_pb2.FcoValidationArgs(
                entity=validator_pb2.EntityValidationArgs(
                    args=self._args,
                )
            )
        else:
            return validator_pb2.FcoValidationArgs(entity=self._spec.data_proto.validation_args)

    @property
    def _is_valid(self) -> bool:
        return self._spec is not None

    @property
    def join_keys(self) -> List[str]:
        """Join keys of the entity."""
        if self._spec is None:
            return list(self._args.join_keys)
        return list(self._spec.join_keys)

    def _validate(self, indentation_level: int = 0) -> None:
        if self._is_valid:
            return

        validations_api.run_backend_validation_and_assert_valid(
            self,
            validator_pb2.ValidationRequest(validation_args=[self._build_fco_validation_args()]),
            indentation_level,
        )

        self._spec = specs.EntitySpec.from_args_proto(self._args)

    @sdk_decorators.sdk_public_method
    @sdk_decorators.assert_remote_object
    def summary(self) -> display.Displayable:
        """Displays a human readable summary of this Feature View."""
        request = metadata_service_pb2.GetEntitySummaryRequest(
            fco_locator=fco_locator_pb2.FcoLocator(id=self._spec.id_proto, workspace=self._spec.workspace)
        )
        response = metadata_service.instance().GetEntitySummary(request)
        return display.Displayable.from_fco_summary(response.fco_summary)

    @sdk_decorators.assert_local_object
    def _create_unvalidated_spec(self) -> specs.EntitySpec:
        """Create an unvalidated spec. Used for user unit testing, where backend validation is unavailable."""
        return specs.EntitySpec.from_args_proto(self._args)
